from django.db import models
from django.urls import reverse
from django.utils.translation import gettext_lazy as _

from mayan.apps.databases.model_mixins import ExtraDataModelMixin
from mayan.apps.documents.models.document_models import Document
from mayan.apps.events.decorators import method_event
from mayan.apps.events.event_managers import EventManagerSave

from .events import event_tag_created, event_tag_edited
from .model_mixins import TagBusinessLogicMixin


class Tag(ExtraDataModelMixin, TagBusinessLogicMixin, models.Model):
    """
    This model represents a binary property that can be applied to a document.
    The tag can have a label and a color.
    """
    _ordering_fields = ('id', 'label')

    label = models.CharField(
        db_index=True, help_text=_(
            message='A short text used as the tag name.'
        ), max_length=128, unique=True, verbose_name=_(message='Label')
    )
    color = models.CharField(
        help_text=_(message='The RGB color values for the tag.'),
        max_length=7, verbose_name=_(message='Color')
    )
    documents = models.ManyToManyField(
        related_name='tags', to=Document, verbose_name=_(message='Documents')
    )

    class Meta:
        ordering = ('label',)
        verbose_name = _(message='Tag')
        verbose_name_plural = _(message='Tags')

    def __str__(self):
        return self.label

    def get_absolute_url(self):
        return reverse(
            kwargs={'tag_id': self.pk}, viewname='tags:tag_document_list'
        )

    @method_event(
        event_manager_class=EventManagerSave,
        created={
            'event': event_tag_created,
            'target': 'self'
        },
        edited={
            'event': event_tag_edited,
            'target': 'self'
        }
    )
    def save(self, *args, **kwargs):
        return super().save(*args, **kwargs)


class DocumentTag(Tag):
    class Meta:
        proxy = True
        verbose_name = _(message='Document tag')
        verbose_name_plural = _(message='Document tags')



from django.utils.html import conditional_escape

from mayan.apps.forms import forms


class TagFormWidget(forms.SelectMultiple):
    def __init__(self, *args, **kwargs):
        return super().__init__(*args, **kwargs)

    def create_option(
        self, name, value, label, selected, index, subindex=None, attrs=None
    ):
        result = super().create_option(
            attrs=attrs, index=index,
            label='{}'.format(
                conditional_escape(text=label)
            ), name=name, selected=selected, subindex=subindex,
            value=value
        )

        result['attrs']['data-color'] = value.instance.color

        return result



from django.urls import re_path

from .api_views import APIMessageDetailView, APIMessageListView
from .views import (
    MessageCreateView, MessageDeleteView, MessageDetailView, MessageListView,
    MessageMarkReadAllView, MessageMarkReadView, MessageMarkUnReadView
)

urlpatterns = [
    re_path(
        route=r'^messages/$', name='message_list',
        view=MessageListView.as_view()
    ),
    re_path(
        route=r'^messages/create/$', name='message_create',
        view=MessageCreateView.as_view()
    ),
    re_path(
        route=r'^messages/(?P<message_id>\d+)/delete/$',
        name='message_single_delete', view=MessageDeleteView.as_view()
    ),
    re_path(
        route=r'^messages/(?P<message_id>\d+)/details/$',
        name='message_detail', view=MessageDetailView.as_view()
    ),
    re_path(
        route=r'^messages/multiple/delete/$',
        name='message_multiple_delete', view=MessageDeleteView.as_view()
    ),
    re_path(
        route=r'^messages/mark_read/$',
        name='message_multiple_mark_read', view=MessageMarkReadView.as_view()
    ),
    re_path(
        route=r'^messages/mark_unread/$',
        name='message_multiple_mark_unread',
        view=MessageMarkUnReadView.as_view()
    ),
    re_path(
        route=r'^messages/(?P<message_id>\d+)/mark_read/$',
        name='message_single_mark_read', view=MessageMarkReadView.as_view()
    ),
    re_path(
        route=r'^messages/(?P<message_id>\d+)/mark_unread/$',
        name='message_single_mark_unread',
        view=MessageMarkUnReadView.as_view()
    ),
    re_path(
        route=r'^messages/all/mark_read/$',
        name='message_all_mark_read',
        view=MessageMarkReadAllView.as_view()
    )
]

api_urls = [
    re_path(
        route=r'^messages/$', name='message-list',
        view=APIMessageListView.as_view()
    ),
    re_path(
        route=r'^messages/(?P<message_id>[0-9]+)/$', name='message-detail',
        view=APIMessageDetailView.as_view()
    )
]



COMMENT_APP_TEMPLATE_CACHE_DISABLE = '{# appearance_app_template_nocache #}'

COMMAND_NAME_APPEARANCE_PREPARE_STATIC = 'appearance_prepare_static'

DEFAULT_APPEARANCE_ELIDED_PAGER_ON_EACH_SIDE = 3
DEFAULT_APPEARANCE_ELIDED_PAGER_ON_ENDS = 1

DEFAULT_APPEARANCE_PAGINATION_DROPDOWN_ENABLE = False
DEFAULT_APPEARANCE_PAGINATION_DROPDOWN_RANGE = 100
DEFAULT_APPEARANCE_PAGINATION_INPUT_ENABLE = False

DEFAULT_AJAX_REDIRECTION_CODE = 278
DEFAULT_MAXIMUM_TITLE_LENGTH = 120
DEFAULT_MESSAGE_POSITION = 'bottom-right'
DEFAULT_MENU_POLLING_INTERVAL = 5000
DEFAULT_THROTTLING_MAXIMUM_REQUESTS = 5
DEFAULT_THROTTLING_TIMEOUT = 4000



from django.utils.translation import gettext_lazy as _

from mayan.apps.documents.models.document_file_models import DocumentFile
from mayan.apps.forms import forms
from mayan.apps.templating.fields import ModelTemplateField

from ..forms import FormDocumentTypeFileMetadataDriverConfiguration


class ViewMixinDynamicConfigurationFormClass:
    def get_form_class(self):
        FormDriverArguments = self.object.stored_driver.driver_class.get_form_class()

        if not FormDriverArguments:
            # Driver does not specify a form and does not have
            # a template either. Create a dynamic form based on the argument
            # list.
            obj = self.object

            argument_values_from_settings = obj.stored_driver.driver_class.get_argument_values_from_settings()

            class FormDriverArguments(forms.Form):
                def __init__(self, *args, **kwargs):
                    super().__init__(*args, **kwargs)
                    argument_name_list = obj.stored_driver.driver_class.get_argument_name_list()

                    for argument_name in argument_name_list:
                        default_value = argument_values_from_settings.get(
                            argument_name
                        )

                        self.fields[argument_name] = ModelTemplateField(
                            initial_help_text=_(
                                message='The template string to be '
                                'evaluated. Leave empty to use the driver\'s '
                                'value passed via settings. Default '
                                'value: %s.'
                            ) % default_value, model=DocumentFile,
                            model_variable='document_file', required=False
                        )

        class FormDriverArgumentsMerged(
            FormDriverArguments,
            FormDocumentTypeFileMetadataDriverConfiguration
        ):
            """Model form merged with the specific transformation fields."""
            view = self

        return FormDriverArgumentsMerged



import datetime
import json
import logging

from django.apps import apps
from django.conf import settings
from django.utils.timezone import now

from mayan.apps.acls.models import AccessControlList

from ..permissions import permission_workflow_instance_transition

logger = logging.getLogger(name=__name__)


class WorkflowInstanceBusinessLogicMixin:
    def do_check_escalation(self):
        current_state = self.get_current_state()

        if not current_state.final and not self.workflow.ignore_completed:
            for escalation in current_state.escalations.filter(enabled=True):
                kwargs = {escalation.unit: escalation.amount}
                timedelta = datetime.timedelta(**kwargs)

                expiration_datetime = self.get_last_log_entry_datetime() + timedelta

                if now() > expiration_datetime:
                    condition_context = {'workflow_instance': self}

                    condition_result = escalation.evaluate_condition(
                        context=condition_context
                    )
                    if condition_result:
                        escalation_comment = escalation.get_comment()

                        self.do_transition(
                            comment=escalation_comment,
                            transition=escalation.transition
                        )

    def do_context_update(self, context):
        workflow_instance_context = self.loads()
        workflow_instance_context.update(context)
        self.dumps(context=workflow_instance_context)

    def do_transition(
        self, transition, comment=None, extra_data=None, user=None
    ):
        comment = comment or ''
        extra_data = extra_data or {}

        try:
            queryset_transitions = self.get_queryset_valid_transitions(user=user).all()

            if queryset_transitions.filter(pk=transition.pk).exists():
                if extra_data:
                    self.do_context_update(context=extra_data)

                return transition.do_execute(
                    comment=comment, extra_data=json.dumps(obj=extra_data),
                    user=user, workflow_instance=self
                )
        except AttributeError:
            # No initial state has been set for this workflow.
            if settings.DEBUG:
                raise

    def dumps(self, context):
        """
        Serialize the context data.
        """
        self.context = json.dumps(obj=context)
        self.save()

    def get_context(self):
        # Keep the document instance in the workflow instance fresh when
        # there are cascade state actions, where a second state action is
        # triggered by the events generated by a first state action.
        self.document.refresh_from_db()
        context = {'workflow_instance': self}
        context['workflow_instance_context'] = self.loads()
        return context

    def get_current_state(self):
        """
        Wrapper for compatibility < 4.9.
        To be removed in a future minor version.
        """
        self.refresh_from_db()

        return self.state_active

    def get_last_log_entry(self):
        return self.log_entries.order_by('datetime').last()

    def get_last_log_entry_datetime(self):
        last_log_entry = self.get_last_log_entry()

        if last_log_entry:
            return last_log_entry.datetime
        else:
            return self.datetime

    def get_last_transition(self):
        """
        Last Transition - The last transition used by the last user to put
        the document in the actual state.
        """
        last_log_entry = self.get_last_log_entry()
        if last_log_entry:
            return last_log_entry.transition

    def get_queryset_valid_transitions(self, user=None):
        current_state = self.get_current_state()

        queryset = current_state.origin_transitions.all()

        if user:
            queryset = AccessControlList.objects.restrict_queryset(
                permission=permission_workflow_instance_transition,
                queryset=queryset, user=user
            )

        # Remove the transitions with a false return value.
        condition_context = {'workflow_instance': self}
        for entry in queryset:
            condition_result = entry.evaluate_condition(
                context=condition_context
            )
            if not condition_result:
                queryset = queryset.exclude(id=entry.pk)

        return queryset

    def get_runtime_context(self):
        """
        Alias of self.load() to get just the runtime context of the instance
        for ease of use in the condition template.
        """
        return self.loads()

    def loads(self):
        """
        Deserialize the context data.
        """
        return json.loads(s=self.context or '{}')


class WorkflowInstanceLogEntryBusinessLogicMixin:
    def get_extra_data(self):
        WorkflowTransitionField = apps.get_model(
            app_label='document_states', model_name='WorkflowTransitionField'
        )

        result = {}
        extra_data = self.loads()

        for key, value in extra_data.items():
            try:
                field = self.transition.fields.get(name=key)
            except WorkflowTransitionField.DoesNotExist:
                """
                There is a reference for a field that does not exist or
                has been deleted.
                """
            else:
                result[field.label] = value

        return result

    def loads(self):
        """
        Deserialize the context data.
        """
        return json.loads(s=self.extra_data or '{}')



from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ('metadata', '0019_alter_documentmetadata_value')
    ]

    operations = [
        migrations.CreateModel(
            bases=('metadata.documentmetadata',), fields=[],
            name='DocumentMetadataSearchResult', options={
                'proxy': True,
                'indexes': [],
                'constraints': []
            }
        )
    ]



from django.contrib import messages
from django.utils.translation import gettext_lazy as _

from actstream.models import Action

from mayan.apps.databases.classes import QuerysetParametersSerializer
from mayan.apps.views.generics import ConfirmView
from mayan.apps.views.view_mixins import ExternalContentTypeObjectViewMixin

from ..icons import (
    icon_event_list_clear, icon_object_event_list_clear,
    icon_verb_event_list_clear
)
from ..permissions import permission_events_clear
from ..tasks import task_event_queryset_clear

from .view_mixins import VerbEventViewMixin


class EventClearBaseView(ConfirmView):
    object_permission = permission_events_clear
    view_icon = icon_event_list_clear

    def get_extra_context(self):
        return {
            'message': _(
                message='This action is not reversible. The process will be '
                'performed in the background. '
            )
        }

    def get_task_extra_kwargs(self):
        return {}

    def view_action(self):
        decomposed_queryset = QuerysetParametersSerializer.decompose(
            _model=Action, **self.get_queryset_parameters()
        )

        task_kwargs = {
            'decomposed_queryset': decomposed_queryset,
            'user_id': self.request.user.pk
        }

        task_kwargs.update(
            self.get_task_extra_kwargs()
        )

        task_event_queryset_clear.apply_async(kwargs=task_kwargs)

        messages.success(
            message=_(
                message='Event list clear task queued successfully.'
            ), request=self.request
        )


class EventListClearView(EventClearBaseView):
    object_permission = permission_events_clear

    def get_extra_context(self):
        context = super().get_extra_context()
        context.update(
            {
                'title': _(message='Clear events')
            }
        )
        return context

    def get_queryset_parameters(self):
        return {'_method_name': 'all'}


class ObjectEventClearView(
    ExternalContentTypeObjectViewMixin, EventClearBaseView
):
    view_icon = icon_object_event_list_clear

    def get_extra_context(self):
        context = super().get_extra_context()
        context.update(
            {
                'object': self.external_object,
                'title': _(
                    message='Clear events of: %s'
                ) % self.external_object
            }
        )
        return context

    def get_queryset_parameters(self):
        return {'_method_name': 'any', 'obj': self.external_object}

    def get_task_extra_kwargs(self):
        return {
            'target_content_type_id': self.external_object_content_type.pk,
            'target_object_id': self.external_object.pk
        }


class VerbEventClearView(VerbEventViewMixin, EventClearBaseView):
    view_icon = icon_verb_event_list_clear

    def get_extra_context(self):
        context = super().get_extra_context()
        context.update(
            {
                'title': _(
                    message='Clear events of type: %s'
                ) % self.event_type
            }
        )
        return context

    def get_queryset_parameters(self):
        return {'_method_name': 'filter', 'verb': self.event_type.id}



from django.utils.translation import gettext_lazy as _

from mayan.apps.dependencies.classes import JavaScriptDependency

JavaScriptDependency(
    label=_(message='JavaScript image cropper'), module=__name__,
    name='cropperjs', version_string='=1.6.2'
)
JavaScriptDependency(
    module=__name__, name='jquery-cropper', version_string='=1.0.2'
)



from collections import defaultdict

from django.db import migrations

from mayan.apps.common.utils import convert_to_internal_name

BULK_UPDATE_SIZE = 100
ITERATOR_CHUNK_SIZE = 1000


def code_populate_internal_name(apps, schema_editor):
    FileMetadataEntry = apps.get_model(
        app_label='file_metadata', model_name='FileMetadataEntry'
    )

    manager_file_metadata_entry = FileMetadataEntry.objects.using(
        alias=schema_editor.connection.alias
    )

    queryset_file_metadata_entry = (
        manager_file_metadata_entry.only(
            'pk', 'document_file_driver_entry_id', 'key', 'internal_name'
        ).order_by(
            'document_file_driver_entry_id', 'key', 'pk'
        )
    )

    document_file_driver_entry_id = None
    list_to_update = []
    counters = defaultdict(int)

    for file_metadata_entry in queryset_file_metadata_entry.iterator(chunk_size=ITERATOR_CHUNK_SIZE):
        if file_metadata_entry.document_file_driver_entry_id != document_file_driver_entry_id:
            document_file_driver_entry_id = file_metadata_entry.document_file_driver_entry_id
            counters.clear()

        internal_name = convert_to_internal_name(
            value=file_metadata_entry.key
        )

        index = counters[internal_name]

        if index == 0:
            internal_name_final = internal_name
        else:
            internal_name_final = f'{internal_name}_{index}'

        counters[internal_name] = index + 1

        file_metadata_entry.internal_name = internal_name_final

        list_to_update.append(file_metadata_entry)

        if len(list_to_update) >= BULK_UPDATE_SIZE:
            manager_file_metadata_entry.bulk_update(
                list_to_update, ['internal_name'], batch_size=BULK_UPDATE_SIZE
            )
            list_to_update.clear()

    if list_to_update:
        manager_file_metadata_entry.bulk_update(
            list_to_update, ['internal_name'], batch_size=BULK_UPDATE_SIZE
        )


class Migration(migrations.Migration):
    dependencies = [
        ('file_metadata', '0010_add_internal_name')
    ]

    operations = [
        migrations.RunPython(
            code=code_populate_internal_name,
            reverse_code=migrations.RunPython.noop
        )
    ]



from django.db import models
from django.utils.translation import gettext_lazy as _

from mayan.apps.common.utils import convert_to_internal_name
from mayan.apps.common.validators import YAMLValidator
from mayan.apps.documents.models.document_file_models import DocumentFile
from mayan.apps.documents.models.document_type_models import DocumentType

from .managers import (
    FileMetadataEntryManager,
    ModelManagerDocumentTypeDriverConfigurationValid,
    ModelManagerStoredDriverValid
)
from .model_mixins import (
    DocumentFileDriverEntryBusinessLogicMixin,
    DocumentTypeDriverConfigurationBusinessLogicMixin,
    FileMetadataEntryBusinessLogicMixin,
    StoredDriverBusinessLogicMixin
)


class DocumentFileDriverEntry(
    DocumentFileDriverEntryBusinessLogicMixin, models.Model
):
    driver = models.ForeignKey(
        on_delete=models.CASCADE, related_name='driver_entries',
        to='StoredDriver', verbose_name=_(message='Driver')
    )
    document_file = models.ForeignKey(
        on_delete=models.CASCADE, related_name='file_metadata_drivers',
        to=DocumentFile, verbose_name=_(message='Document file')
    )

    class Meta:
        ordering = ('document_file', 'driver')
        unique_together = ('driver', 'document_file')
        verbose_name = _(message='Document file driver entry')
        verbose_name_plural = _(message='Document file driver entries')

    def __str__(self):
        return str(self.driver)


class DocumentTypeDriverConfiguration(
    DocumentTypeDriverConfigurationBusinessLogicMixin, models.Model
):
    _ordering_fields = ('enabled',)

    arguments = models.TextField(
        blank=True, help_text=_(
            message='Enter the arguments for the drive for the specific '
            'document type as a YAML dictionary. ie: {"degrees": 180}'
        ), validators=[
            YAMLValidator()
        ], verbose_name=_(message='Arguments')
    )
    document_type = models.ForeignKey(
        on_delete=models.CASCADE,
        related_name='file_metadata_driver_configurations',
        to=DocumentType, verbose_name=_(message='Document type')
    )
    stored_driver = models.ForeignKey(
        on_delete=models.CASCADE,
        related_name='document_type_configurations',
        to='StoredDriver', verbose_name=_(message='Driver')
    )
    enabled = models.BooleanField(
        default=True, help_text=_(
            'Enable this driver to process document files of the selected '
            'document type.'
        ), verbose_name=_(message='Enabled')
    )

    objects = models.Manager()
    valid = ModelManagerDocumentTypeDriverConfigurationValid()

    def __str__(self):
        return str(self.stored_driver)

    class Meta:
        ordering = ('stored_driver',)
        unique_together = ('document_type', 'stored_driver')
        verbose_name = _(message='Document type driver settings')
        verbose_name_plural = _(message='Document type driver settings')


class FileMetadataEntry(FileMetadataEntryBusinessLogicMixin, models.Model):
    _ordering_fields = ('internal_name', 'key', 'value')

    document_file_driver_entry = models.ForeignKey(
        on_delete=models.CASCADE, related_name='entries',
        to=DocumentFileDriverEntry,
        verbose_name=_(message='Document file driver entry')
    )
    internal_name = models.CharField(
        db_index=True, help_text=_(
            message='Normalized name of the file metadata entry.'
        ), max_length=255, verbose_name=_(message='Internal name')
    )
    key = models.CharField(
        help_text=_(
            message='Name of the file metadata entry as provided by the '
            'driver.'
        ), max_length=255, verbose_name=_(message='Key')
    )
    value = models.TextField(
        blank=True, help_text=_(message='Value of the file metadata entry.'),
        max_length=255, verbose_name=_(message='Value')
    )

    class Meta:
        ordering = ('internal_name', 'value')
        unique_together = ('document_file_driver_entry', 'internal_name')
        verbose_name = _(message='File metadata entry')
        verbose_name_plural = _(message='File metadata entries')

    objects = FileMetadataEntryManager()

    def __str__(self):
        return '{}: {}: {}'.format(
            self.document_file_driver_entry, self.key, self.value
        )

    def save(self, *args, **kwargs):
        internal_name = convert_to_internal_name(value=self.key)

        queryset_siblings = self.document_file_driver_entry.entries.exclude(
            pk=self.pk
        )

        queryset_duplicated = queryset_siblings.filter(
            internal_name=internal_name
        )

        if queryset_duplicated.exists():
            internal_name = '{}_{}'.format(
                internal_name, queryset_duplicated.count()
            )

        self.internal_name = internal_name

        super().save(*args, **kwargs)


class StoredDriver(StoredDriverBusinessLogicMixin, models.Model):
    _ordering_fields = ('internal_name',)

    driver_path = models.CharField(
        max_length=255, unique=True, verbose_name=_(message='Driver path')
    )
    internal_name = models.CharField(
        db_index=True, max_length=128, unique=True,
        verbose_name=_(message='Internal name')
    )
    exists = models.BooleanField(
        default=True, help_text=_(
            'The class defined by this instance is valid and active.'
        ), verbose_name=_(message='Valid')
    )

    objects = models.Manager()
    valid = ModelManagerStoredDriverValid()

    class Meta:
        ordering = ('internal_name',)
        verbose_name = _(message='Driver')
        verbose_name_plural = _(message='Drivers')

    def __str__(self):
        return str(self.driver_label)



from django.utils.translation import gettext_lazy as _

from mayan.apps.documents.models.document_version_models import (
    DocumentVersion
)
from mayan.apps.organizations.utils import get_organization_installation_url
from mayan.apps.views.generics import MultipleObjectConfirmActionView

from .icons import icon_document_version_export
from .permissions import permission_document_version_export
from .tasks import task_document_version_export


class DocumentVersionExportView(MultipleObjectConfirmActionView):
    object_permission = permission_document_version_export
    pk_url_kwarg = 'document_version_id'
    source_queryset = DocumentVersion.valid.all()
    success_message_single = _(
        message='Document version "%(object)s" export successfully queued.'
    )
    success_message_singular = _(
        message='%(count)d document version export successfully queued.'
    )
    success_message_plural = _(
        message='%(count)d document versions exports successfully queued.'
    )
    title_plural = _(message='Export %(count)d document versions.')
    title_single = _(message='Export document version "%(object)s".')
    title_singular = _(message='Export %(count)d document version.')
    view_icon = icon_document_version_export

    def get_extra_context(self):
        context = {
            'message': _(
                message='The process will be performed in the background. '
                'The exported file will be available in the downloads area.'
            )
        }

        if self.object_list.count() == 1:
            context['object'] = self.object_list.first()

        return context

    def object_action(self, form, instance):
        task_document_version_export.apply_async(
            kwargs={
                'document_version_id': instance.pk,
                'organization_installation_url': get_organization_installation_url(
                    request=self.request
                ),
                'user_id': self.request.user.pk
            }
        )



from mayan.apps.forms import form_fields, forms

from ..classes import Layer
from ..forms import LayerTransformationForm


class ViewMixinDynamicTransformationFormClass:
    def get_form_class(self):
        transformation_class = self.get_transformation_class()

        TransformationForm = transformation_class.get_form_class()

        transformation_has_form_or_template = TransformationForm or transformation_class.get_template_name()

        if not transformation_has_form_or_template:
            # Transformation does not specify a form and does not have
            # a template either. Create a dynamic form based on the argument
            # list.
            class TransformationForm(forms.Form):
                def __init__(self, *args, **kwargs):
                    super().__init__(*args, **kwargs)
                    for argument in transformation_class.get_arguments():
                        self.fields[argument] = form_fields.CharField(
                            required=False
                        )

        class MergedTransformationForm(
            TransformationForm, LayerTransformationForm
        ):
            """Model form merged with the specific transformation fields."""
            view = self

        return MergedTransformationForm


class ViewMixinLayer:
    def dispatch(self, request, *args, **kwargs):
        self.layer = self.get_layer()
        return super().dispatch(request=request, *args, **kwargs)

    def get_layer(self):
        return Layer.get(
            name=self.kwargs['layer_name']
        )


class ViewMixinTransformationTemplateName:
    def get_template_names(self):
        transformation_template_name = self.get_transformation_template_name()

        if transformation_template_name is None:
            transformation_template_name = self.template_name

        return (transformation_template_name,)

    def get_transformation_template_name(self):
        transformation_class = self.get_transformation_class()

        return transformation_class.get_template_name()



from django.http import HttpResponseRedirect
from django.utils.translation import gettext_lazy as _
from django.views.generic.base import View

from mayan.apps.common.settings import setting_home_view
from mayan.apps.user_management.querysets import get_user_queryset
from mayan.apps.views.generics import FormView
from mayan.apps.views.http import URL
from mayan.apps.views.view_mixins import (
    ExternalObjectViewMixin, RedirectionViewMixin
)

from ..forms import (
    UserImpersonationOptionsForm, UserImpersonationSelectionForm
)
from ..icons import icon_impersonate_start
from ..literals import (
    USER_IMPERSONATE_VARIABLE_DISABLE, USER_IMPERSONATE_VARIABLE_ID,
    USER_IMPERSONATE_VARIABLE_PERMANENT
)
from ..permissions import permission_users_impersonate


class UserImpersonateEndView(RedirectionViewMixin, View):
    def get(self, request, *args, **kwargs):
        url = URL(
            viewname=setting_home_view.value, query={
                USER_IMPERSONATE_VARIABLE_DISABLE: ''
            }
        )
        return HttpResponseRedirect(
            redirect_to=url.to_string()
        )


class UserImpersonateFormStartView(FormView):
    form_class = UserImpersonationSelectionForm
    view_icon = icon_impersonate_start

    def form_valid(self, form):
        query = {
            USER_IMPERSONATE_VARIABLE_ID: form.cleaned_data['user_to_impersonate'].pk
        }
        if form.cleaned_data['permanent']:
            query[USER_IMPERSONATE_VARIABLE_PERMANENT] = ''

        url = URL(
            viewname=setting_home_view.value, query=query
        )
        return HttpResponseRedirect(
            redirect_to=url.to_string()
        )

    def get_extra_context(self):
        return {
            'title': _(message='Impersonate user')
        }

    def get_form_extra_kwargs(self):
        return {
            'user': self.request.user
        }


class UserImpersonateStartView(ExternalObjectViewMixin, FormView):
    external_object_queryset = get_user_queryset()
    external_object_permission = permission_users_impersonate
    external_object_pk_url_kwarg = 'user_id'
    form_class = UserImpersonationOptionsForm
    view_icon = icon_impersonate_start

    def form_valid(self, form):
        query = {USER_IMPERSONATE_VARIABLE_ID: self.external_object.pk}
        if form.cleaned_data['permanent']:
            query[USER_IMPERSONATE_VARIABLE_PERMANENT] = ''

        url = URL(
            query=query, viewname=setting_home_view.value
        )
        return HttpResponseRedirect(
            redirect_to=url.to_string()
        )

    def get_extra_context(self):
        return {
            'object': self.external_object,
            'title': _(message='Impersonate user: %s') % self.external_object
        }



from django.urls import re_path

from .api_views import (
    APIMailObjectAttachmentView, APIMailObjectLinkView,
    APIMailingProfileDetailView, APIMailingProfileListView
)
from .views.document_file_views import (
    MailDocumentFileAttachmentView, MailDocumentFileLinkView
)
from .views.document_version_views import (
    MailDocumentVersionAttachmentView, MailDocumentVersionLinkView
)
from .views.document_views import MailDocumentLinkView
from .views.mailing_profile_views import (
    MailingProfileBackendSelectionView, MailingProfileCreateView,
    MailingProfileDeleteView, MailingProfileEditView, MailingProfileListView,
    MailingProfileTestView
)

urlpatterns_document = [
    re_path(
        route=r'^documents/(?P<document_id>\d+)/send/link/$',
        name='send_document_link_single',
        view=MailDocumentLinkView.as_view()
    ),
    re_path(
        route=r'^documents/multiple/send/link/$',
        name='send_document_link_multiple',
        view=MailDocumentLinkView.as_view()
    )
]

urlpatterns_document_file = [
    re_path(
        route=r'^documents/files/(?P<document_file_id>\d+)/send/attachment/$',
        name='send_document_file_attachment_single',
        view=MailDocumentFileAttachmentView.as_view()
    ),
    re_path(
        route=r'^documents/files/multiple/send/attachment/$',
        name='send_document_file_attachment_multiple',
        view=MailDocumentFileAttachmentView.as_view()
    ),
    re_path(
        route=r'^documents/files/(?P<document_file_id>\d+)/send/link/$',
        name='send_document_file_link_single',
        view=MailDocumentFileLinkView.as_view()
    ),
    re_path(
        route=r'^documents/files/multiple/send/link/$',
        name='send_document_file_link_multiple',
        view=MailDocumentFileLinkView.as_view()
    )
]

urlpatterns_document_version = [
    re_path(
        route=r'^documents/versions/(?P<document_version_id>\d+)/send/attachment/$',
        name='send_document_version_attachment_single',
        view=MailDocumentVersionAttachmentView.as_view()
    ),
    re_path(
        route=r'^documents/versions/multiple/send/attachment/$',
        name='send_document_version_attachment_multiple',
        view=MailDocumentVersionAttachmentView.as_view()
    ),
    re_path(
        route=r'^documents/versions/(?P<document_version_id>\d+)/send/link/$',
        name='send_document_version_link_single',
        view=MailDocumentVersionLinkView.as_view()
    ),
    re_path(
        route=r'^documents/versions/multiple/send/link/$',
        name='send_document_version_link_multiple',
        view=MailDocumentVersionLinkView.as_view()
    )
]

urlpatterns_mailing_profiles = [
    re_path(
        route=r'^mailing_profiles/backend/selection/$',
        name='mailing_profile_backend_selection',
        view=MailingProfileBackendSelectionView.as_view()
    ),
    re_path(
        route=r'^mailing_profiles/(?P<backend_path>[a-zA-Z0-9_.]+)/create/$',
        name='mailing_profile_create', view=MailingProfileCreateView.as_view()
    ),
    re_path(
        route=r'^mailing_profiles/(?P<mailing_profile_id>\d+)/delete/$',
        name='mailing_profile_delete', view=MailingProfileDeleteView.as_view()
    ),
    re_path(
        route=r'^mailing_profiles/(?P<mailing_profile_id>\d+)/edit/$',
        name='mailing_profile_edit', view=MailingProfileEditView.as_view()
    ),
    re_path(
        route=r'^mailing_profiles/(?P<mailing_profile_id>\d+)/test/$',
        name='mailing_profile_test', view=MailingProfileTestView.as_view()
    ),
    re_path(
        route=r'^mailing_profiles/$', name='mailing_profile_list',
        view=MailingProfileListView.as_view()
    )
]

urlpatterns = []
urlpatterns.extend(urlpatterns_document)
urlpatterns.extend(urlpatterns_document_file)
urlpatterns.extend(urlpatterns_document_version)
urlpatterns.extend(urlpatterns_mailing_profiles)

api_urls_mailing_profiles = [
    re_path(
        route=r'^mailing_profiles/$', name='mailing_profile-list',
        view=APIMailingProfileListView.as_view()
    ),
    re_path(
        route=r'^mailing_profiles/(?P<mailing_profile_id>[0-9]+)/$',
        name='mailing_profile-detail',
        view=APIMailingProfileDetailView.as_view()
    )
]

api_urls_mailing_profile_actions = [
    re_path(
        route=r'^objects/(?P<app_label>[-\w]+)/(?P<model_name>[-\w]+)/(?P<object_id>\d+)/mailing/attachment/$',
        name='object-mailing-action-attachment',
        view=APIMailObjectAttachmentView.as_view()
    ),
    re_path(
        route=r'^objects/(?P<app_label>[-\w]+)/(?P<model_name>[-\w]+)/(?P<object_id>\d+)/mailing/link/$',
        name='object-mailing-action-link', view=APIMailObjectLinkView.as_view()
    )
]

api_urls = []
api_urls.extend(api_urls_mailing_profiles)
api_urls.extend(api_urls_mailing_profile_actions)



from openai import OpenAI

from django.utils.translation import gettext_lazy as _

from mayan.apps.common.serialization import yaml_load
from mayan.apps.common.utils import flatten_object
from mayan.apps.file_metadata.classes import FileMetadataDriver
from mayan.apps.file_metadata.literals import RESULT_SEPARATOR
from mayan.apps.forms import forms
from mayan.apps.templating.fields import TemplateField
from mayan.apps.templating.template_backends import Template

from .literals import DEFAULT_TIMEOUT


class FileMetadataDriverOpenAIResponseAPI(FileMetadataDriver):
    argument_name_list = (
        'api_key', 'base_url', 'input', 'model', 'organization', 'project',
        'timeout'
    )
    description = _(message='Analyze content using OpenAI\'s Response API.')
    enabled = False
    internal_name = 'openai_response'
    label = _(message='OpenAI Response API driver')
    mime_type_list = ('*',)

    class Form(forms.Form):
        class FormMeta:
            fieldsets = (
                (
                    _(message='Basic'), {
                        'fields': ('enabled', 'arguments')
                    }
                ),
                (
                    _(message='Authentication'), {
                        'fields': ('api_key',)
                    }
                ),
                (
                    _(message='Control'), {
                        'fields': ('organization', 'project', 'model')
                    }
                ),
                (
                    _(message='Prompt'), {
                        'fields': ('input',)
                    }
                ),
                (
                    _(message='Communication'), {
                        'fields': ('base_url', 'timeout',)
                    }
                ),
            )

        base_url = TemplateField(
            initial_help_text=_(
                message='API endpoint base URL. Change this only when using '
                'a compatible non-default endpoint or proxy. The document '
                'file object is available as {{ document_file }}.'
            ), label=_(message='Base URL'), required=False
        )
        api_key = TemplateField(
            initial_help_text=_(
                message='The API key used for authentication. The document '
                'file object is available as {{ document_file }}.'
            ), label=_(message='API Key'), required=True
        )
        input = TemplateField(
            initial_help_text=_(
                message='Text or file content to send to the OpenAI API '
                'for analysis. The document file object is available as '
                '{{ document_file }}.'
            ), label=_(message='Input'), required=False
        )
        model = TemplateField(
            initial_help_text=_(
                message='Identifier of the OpenAI model to use '
                '(for example: gpt-5, gpt-5-nano). The document file object '
                'is available as {{ document_file }}.'
            ), label=_(message='Model'), required=True
        )
        organization = TemplateField(
            initial_help_text=_(
                message='Optional OpenAI organization ID. Used when the '
                'API key is linked to multiple organizations. The document '
                'file object is available as {{ document_file }}.'
            ), label=_(message='Organization'), required=False
        )
        project = TemplateField(
            initial_help_text=_(
                message='Optional OpenAI project ID. Allows scoping '
                'requests to a specific project.  The document file object '
                'is available as {{ document_file }}.'
            ), label=_(message='Project'), required=False
        )
        timeout = TemplateField(
            initial_help_text=_(
                message='Maximum time (in seconds) to wait for a response '
                'from the API before failing.  The document file object is '
                'available as {{ document_file }}.'
            ), label=_(message='Timeout'), required=False
        )

    @classmethod
    def get_argument_values_for_document_file(cls, document_file):
        document_type = document_file.document.document_type

        configuration_instance = document_type.file_metadata_driver_configurations.get(
            stored_driver__internal_name=cls.internal_name
        )

        document_type_arguments = configuration_instance.get_arguments()

        context = {'document_file': document_file}

        argument_values_rendered = {}

        for key, value in document_type_arguments.items():
            template = Template(template_string=value)
            template_result = template.render(context=context)
            argument_values_rendered[key] = template_result

        return argument_values_rendered

    @classmethod
    def get_argument_values_from_settings(cls):
        result = {'timeout': DEFAULT_TIMEOUT}
        setting_arguments = super().get_argument_values_from_settings()

        if setting_arguments:
            result.update(setting_arguments)

        return result

    def __init__(
        self, api_key, input, model, base_url=None, organization=None,
        project=None, timeout=None, **kwargs
    ):
        super().__init__(**kwargs)

        self.api_key = api_key
        self.base_url = base_url
        self.input = yaml_load(stream=input)
        self.model = model
        self.organization = organization
        self.project = project
        self.timeout = int(timeout)

    def _process(self, document_file):
        result = {}

        client = OpenAI(
            api_key=self.api_key, base_url=self.base_url,
            organization=self.organization, project=self.project
        )

        response = client.responses.create(
            input=self.input, model=self.model
        )

        dictionary = response.to_dict()
        generator = flatten_object(obj=dictionary, separator=RESULT_SEPARATOR)
        result = dict(generator)

        return result



from rest_framework import status
from rest_framework.generics import get_object_or_404
from rest_framework.response import Response

from mayan.apps.documents.api_views.api_view_mixins import (
    ParentObjectDocumentVersionPageAPIViewMixin
)
from mayan.apps.documents.models.document_models import Document
from mayan.apps.documents.models.document_version_models import (
    DocumentVersion
)
from mayan.apps.rest_api import generics

from .models import DocumentVersionPageOCRContent, DocumentTypeOCRSettings
from .permissions import (
    permission_document_type_ocr_setup, permission_document_version_ocr,
    permission_document_version_ocr_content_edit,
    permission_document_version_ocr_content_view
)
from .serializers import (
    DocumentVersionPageOCRContentSerializer, DocumentTypeOCRSettingsSerializer
)


class APIDocumentTypeOCRSettingsView(generics.RetrieveUpdateAPIView):
    """
    get: Return the document type OCR settings.
    patch: Set the document type OCR settings.
    put: Set the document type OCR settings.
    """
    lookup_field = 'document_type__pk'
    lookup_url_kwarg = 'document_type_id'
    mayan_object_permission_map = {
        'GET': permission_document_type_ocr_setup,
        'PATCH': permission_document_type_ocr_setup,
        'PUT': permission_document_type_ocr_setup
    }
    serializer_class = DocumentTypeOCRSettingsSerializer
    source_queryset = DocumentTypeOCRSettings.objects.all()


class APIDocumentOCRSubmitView(generics.GenericAPIView):
    """
    post: Submit a document for OCR.
    """
    lookup_url_kwarg = 'document_id'
    mayan_object_permission_map = {'POST': permission_document_version_ocr}
    source_queryset = Document.valid.all()

    def get_serializer(self, *args, **kwargs):
        return None

    def get_serializer_class(self):
        return None

    def post(self, request, *args, **kwargs):
        self.get_object().submit_for_ocr(user=self.request.user)
        return Response(status=status.HTTP_202_ACCEPTED)


class APIDocumentVersionPageOCRContentDetailView(
    ParentObjectDocumentVersionPageAPIViewMixin,
    generics.RetrieveUpdateAPIView
):
    """
    get: Returns the OCR content of the selected document page.
    patch: Edit the OCR content of the selected document page.
    put: Edit the OCR content of the selected document page.
    """
    mayan_object_permission_map = {
        'GET': permission_document_version_ocr_content_view,
        'PATCH': permission_document_version_ocr_content_edit,
        'PUT': permission_document_version_ocr_content_edit
    }
    serializer_class = DocumentVersionPageOCRContentSerializer

    def get_instance_extra_data(self):
        return {'_event_actor': self.request.user}

    def get_object(self):
        document_version_page = self.get_document_version_page(
            permission=self.mayan_object_permission_map.get(
                self.request.method, None
            )
        )

        try:
            return document_version_page.ocr_content
        except DocumentVersionPageOCRContent.DoesNotExist:
            return DocumentVersionPageOCRContent(
                document_version_page=document_version_page
            )


class APIDocumentVersionOCRSubmitView(generics.GenericAPIView):
    """
    post: Submit a document version for OCR.
    """
    lookup_url_kwarg = 'document_version_id'
    mayan_object_permission_map = {'POST': permission_document_version_ocr}
    source_queryset = DocumentVersion.objects.all()

    def get_document(self):
        return get_object_or_404(
            queryset=Document.valid.all(), pk=self.kwargs['document_id']
        )

    def get_serializer(self, *args, **kwargs):
        return None

    def get_serializer_class(self):
        return None

    def get_source_queryset(self):
        return self.get_document().versions.all()

    def post(self, request, *args, **kwargs):
        self.get_object().submit_for_ocr(user=self.request.user)
        return Response(status=status.HTTP_202_ACCEPTED)



from django.utils.translation import gettext_lazy as _

from mayan.apps.documents.models.document_file_models import DocumentFile
from mayan.apps.documents.models.document_models import Document
from mayan.apps.documents.views.document_file_views import (
    DocumentFileListView
)
from mayan.apps.organizations.utils import get_organization_installation_url
from mayan.apps.storage.views.download_views import (
    ViewSingleObjectBackendDownload
)
from mayan.apps.views.generics import MultipleObjectFormActionView
from mayan.apps.views.view_mixins import MultipleExternalObjectViewMixin

from .events import event_document_file_downloaded
from .forms import DocumentDownloadFormSet
from .icons import (
    icon_document_download_multiple, icon_document_file_download_quick
)
from .permissions import permission_document_file_download
from .settings import (
    setting_document_file_download_backend,
    setting_document_file_download_backend_arguments
)
from .tasks import task_document_file_compress


class DocumentDownloadView(
    MultipleExternalObjectViewMixin, MultipleObjectFormActionView
):
    error_message = _(
        message='Unable to queued document "%(instance)s" for file download; '
        '%(exception)s.'
    )
    external_object_pk_url_kwarg = 'document_id'
    external_object_queryset = Document.valid.all()
    form_class = DocumentDownloadFormSet
    object_permission = permission_document_file_download
    source_queryset = DocumentFile.valid.all()
    success_message_plural = _(
        message='%(count)d document files queued for download.'
    )
    success_message_single = _(
        message='Document file "%(object)s" queued for download.'
    )
    success_message_singular = _(
        message='%(count)d document file queued for download.'
    )
    title_plural = _(message='Download files of %(count)d documents')
    title_single = _(message='Download files of document: %(object)s')
    title_singular = _(message='Download files of %(count)d document')
    view_icon = icon_document_download_multiple

    def get_extra_context(self):
        context = {
            'form_display_mode_table': True,
            'subtitle': _(
                message='The process will be performed in the background. The '
                'document files will be available in the downloads area.'
            )
        }

        if self.object_list.count() == 1:
            document = self.object_list.first().document
            context.update(
                {
                    'object': document
                }
            )
        else:
            document = None

        context.update(
            DocumentFileListView.get_no_results_context(
                document=document, request=self.request
            )
        )

        return context

    def get_initial(self):
        initial = []

        for document_file in self.object_list:
            initial.append(
                {
                    'document': str(document_file.document),
                    'document_file': str(document_file),
                    'document_file_id': document_file.pk
                }
            )

        return initial

    def get_object_list(self):
        return self.get_queryset().filter(
            document_id__in=self.external_object_list.values('pk')
        )

    def object_action(self, instance, form=None):
        """
        There are no actions for individual objects. This methods still needs
        to exists to allow the super call to view_actions to complete.
        """

    def view_action(self, form=None):
        id_list = [
            form_data['document_file_id'] for form_data in form.cleaned_data if form_data['include']
        ]

        queryset = self.get_object_list().filter(pk__in=id_list)

        task_document_file_compress.apply_async(
            kwargs={
                'id_list': list(
                    queryset.values_list('pk', flat=True)
                ),
                'organization_installation_url': get_organization_installation_url(
                    request=self.request
                ),
                'user_id': self.request.user.pk
            }
        )

        return super().view_action(form=form)


class DocumentFileDownloadView(ViewSingleObjectBackendDownload):
    backend_arguments = setting_document_file_download_backend_arguments.value
    backend_path = setting_document_file_download_backend.value
    download_event_type = event_document_file_downloaded
    object_permission = permission_document_file_download
    pk_url_kwarg = 'document_file_id'
    source_queryset = DocumentFile.valid.all()
    view_icon = icon_document_file_download_quick

    def get_download_event_action_object(self):
        return self.object.document



from django.urls import re_path

from .views import (
    UserConfirmViewDeleteView, UserConfirmViewListView, UserViewModeView
)

urlpatterns = [
    re_path(
        route=r'^users/(?P<user_id>\d+)/views/confirm/$',
        name='user_views_confirm_list', view=UserConfirmViewListView.as_view()
    ),
    re_path(
        route=r'^users/views/confirm/(?P<user_confirmation_view_id>\d+)/delete/$',
        name='user_views_confirm_delete_single',
        view=UserConfirmViewDeleteView.as_view()
    ),
    re_path(
        route=r'^users/views/confirm/multiple/delete/$',
        name='user_views_confirm_delete_multiple',
        view=UserConfirmViewDeleteView.as_view()
    ),
    re_path(
        route=r'^users/(?P<user_id>\d+)/views/modes/$',
        name='user_view_modes', view=UserViewModeView.as_view()
    )
]



from mayan.apps.forms import forms

from ..fields import DocumentFilePageField


class DocumentFilePageForm(forms.Form):
    document_file_page = DocumentFilePageField()

    def __init__(self, *args, **kwargs):
        instance = kwargs.pop('instance', None)
        transformation_instance_list = kwargs.pop(
            'transformation_instance_list', ()
        )
        super().__init__(*args, **kwargs)
        self.fields['document_file_page'].initial = instance
        self.fields['document_file_page'].widget.attrs.update(
            {
                'transformation_instance_list': transformation_instance_list
            }
        )



from mayan.apps.rest_api import generics
from mayan.apps.rest_api.api_view_mixins import (
    APIViewMixinExternalObjectOwnerPlusFilteredQueryset,
    APIViewMixinOwnerPlusFilteredQueryset, ExternalObjectAPIViewMixin
)

from ..models import SavedResultset
from ..permissions import (
    permission_saved_resultset_delete, permission_saved_resultset_view
)
from ..serializers import (
    DummySearchResultModelSerializer, SavedResultsetSerializer
)
from ..views.view_mixins import SearchResultViewMixin

from .api_view_mixins import SearchModelAPIViewMixin


class APISavedResultsetCreateView(
    SearchResultViewMixin, SearchModelAPIViewMixin, generics.CreateAPIView
):
    """
    post: Create a saved resultset.
    """
    serializer_class = SavedResultsetSerializer


class APISavedResultsetDetailView(
    APIViewMixinOwnerPlusFilteredQueryset, generics.RetrieveDestroyAPIView
):
    """
    delete: Delete the selected saved resultset.
    get: Return the details of the selected saved resultset.
    """
    lookup_url_kwarg = 'saved_resultset_id'
    mayan_object_permission_map = {
        'DELETE': permission_saved_resultset_delete,
        'GET': permission_saved_resultset_view
    }
    serializer_class = SavedResultsetSerializer
    source_queryset = SavedResultset.objects.all()


class APISavedResultsetListView(
    APIViewMixinOwnerPlusFilteredQueryset, generics.ListAPIView
):
    """
    get: Returns a list of all the saved resultsets.
    """
    mayan_object_permission_map = {'GET': permission_saved_resultset_view}
    serializer_class = SavedResultsetSerializer
    source_queryset = SavedResultset.objects.all()


class APISavedResultsetResultListView(
    APIViewMixinExternalObjectOwnerPlusFilteredQueryset,
    ExternalObjectAPIViewMixin, SearchResultViewMixin, generics.ListAPIView
):
    """
    get: Return the results of the selected saved resultset.
    """
    external_object_class = SavedResultset
    external_object_pk_url_kwarg = 'saved_resultset_id'
    mayan_object_permission_map = {
        'GET': permission_saved_resultset_view
    }

    def get_serializer_class(self):
        if getattr(self, 'swagger_fake_view', False):
            return DummySearchResultModelSerializer
        else:
            external_object = self.get_external_object()
            search_model = external_object.get_search_model()
            return search_model.serializer

    def get_source_queryset(self):
        external_object = self.get_external_object()
        return external_object.get_result_queryset()



from django.template import RequestContext
from django.urls import reverse_lazy
from django.utils.translation import gettext_lazy as _

from mayan.apps.documents.permissions import permission_document_view
from mayan.apps.documents.views.document_views import DocumentListView
from mayan.apps.views.generics import (
    SingleObjectCreateView, SingleObjectDeleteView, SingleObjectEditView,
    SingleObjectListView
)
from mayan.apps.views.view_mixins import ExternalObjectViewMixin

from ..icons import (
    icon_cabinet, icon_cabinet_child_add, icon_cabinet_create,
    icon_cabinet_delete, icon_cabinet_detail, icon_cabinet_edit,
    icon_cabinet_list
)
from ..links import link_cabinet_child_add, link_cabinet_create
from ..models import Cabinet
from ..permissions import (
    permission_cabinet_create, permission_cabinet_delete,
    permission_cabinet_edit, permission_cabinet_view
)
from ..widgets import jstree_data


class CabinetCreateView(SingleObjectCreateView):
    fields = ('label',)
    model = Cabinet
    post_action_redirect = reverse_lazy(viewname='cabinets:cabinet_list')
    view_icon = icon_cabinet_create
    view_permission = permission_cabinet_create

    def get_extra_context(self):
        return {
            'title': _(message='Create cabinet')
        }

    def get_instance_extra_data(self):
        return {'_event_actor': self.request.user}


class CabinetChildAddView(ExternalObjectViewMixin, SingleObjectCreateView):
    fields = ('label',)
    external_object_class = Cabinet
    external_object_permission = permission_cabinet_create
    external_object_pk_url_kwarg = 'cabinet_id'
    view_icon = icon_cabinet_child_add

    def get_extra_context(self):
        return {
            'title': _(
                message='Add new level to: %s'
            ) % self.external_object.get_full_path(),
            'object': self.external_object
        }

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user,
            'parent': self.external_object
        }

    def get_queryset(self):
        return self.external_object.get_descendants()


class CabinetDeleteView(SingleObjectDeleteView):
    model = Cabinet
    object_permission = permission_cabinet_delete
    post_action_redirect = reverse_lazy(viewname='cabinets:cabinet_list')
    pk_url_kwarg = 'cabinet_id'
    view_icon = icon_cabinet_delete

    def get_extra_context(self):
        return {
            'object': self.object,
            'title': _(message='Delete the cabinet: %s?') % self.object
        }


class CabinetDetailView(ExternalObjectViewMixin, DocumentListView):
    external_object_class = Cabinet
    external_object_permission = permission_cabinet_view
    external_object_pk_url_kwarg = 'cabinet_id'
    template_name = 'cabinets/cabinet_details.html'
    view_icon = icon_cabinet_detail

    def get_document_queryset(self):
        return self.external_object.get_documents(
            permission=permission_document_view, user=self.request.user
        )

    def get_extra_context(self, **kwargs):
        context = super().get_extra_context(**kwargs)

        context.update(
            {
                'column_class': 'col-xs-12 col-sm-6 col-md-4 col-lg-3',
                'hide_links': True,
                'jstree_data': '\n'.join(
                    jstree_data(
                        node=self.external_object.get_root(),
                        selected_node=self.external_object
                    )
                ),
                'list_as_items': True,
                'no_results_icon': icon_cabinet,
                'no_results_main_link': link_cabinet_child_add.resolve(
                    context=RequestContext(
                        dict_={
                            'object': self.external_object
                        }, request=self.request
                    )
                ),
                'no_results_text': _(
                    message='Cabinet levels can contain documents or other '
                    'cabinet sub levels. To add documents to a cabinet, '
                    'select the cabinet view of a document view.'
                ),
                'no_results_title': _(message='This cabinet level is empty'),
                'object': self.external_object,
                'title': _(
                    message='Details of cabinet: %s'
                ) % self.external_object.get_full_path()
            }
        )

        return context


class CabinetEditView(SingleObjectEditView):
    fields = ('label',)
    model = Cabinet
    object_permission = permission_cabinet_edit
    post_action_redirect = reverse_lazy(viewname='cabinets:cabinet_list')
    pk_url_kwarg = 'cabinet_id'
    view_icon = icon_cabinet_edit

    def get_extra_context(self):
        return {
            'object': self.object,
            'title': _(message='Edit cabinet: %s') % self.object
        }

    def get_instance_extra_data(self):
        return {'_event_actor': self.request.user}


class CabinetListView(SingleObjectListView):
    object_permission = permission_cabinet_view
    view_icon = icon_cabinet_list

    def get_extra_context(self):
        return {
            'hide_link': True,
            'hide_object': True,
            'title': _(message='Cabinets'),
            'no_results_icon': icon_cabinet,
            'no_results_main_link': link_cabinet_create.resolve(
                context=RequestContext(request=self.request)
            ),
            'no_results_text': _(
                message='Cabinets are a multi-level method to organize '
                'documents. Each cabinet can contain documents as '
                'well as other sub level cabinets.'
            ),
            'no_results_title': _(message='No cabinets available')
        }

    def get_source_queryset(self):
        return Cabinet.objects.root_nodes()



from django.urls import re_path

from .api_views import (
    APIObjectTemplateSandboxActionView, APITemplateDetailView, APITemplateListView
)

from .views import ObjectTemplateSandboxView

urlpatterns = [
    re_path(
        route=r'^objects/(?P<app_label>[-\w]+)/(?P<model_name>[-\w]+)/(?P<object_id>\d+)/sandbox/$',
        name='object_template_sandbox',
        view=ObjectTemplateSandboxView.as_view()
    )
]

api_urls = [
    re_path(
        route=r'^objects/(?P<app_label>[-\w]+)/(?P<model_name>[-\w]+)/(?P<object_id>\d+)/sandbox/$',
        view=APIObjectTemplateSandboxActionView.as_view(),
        name='object-template-sandbox'
    ),
    re_path(
        route=r'^templates/$', view=APITemplateListView.as_view(),
        name='template-list'
    ),
    re_path(
        route=r'^templates/(?P<name>[-\w]+)/$',
        view=APITemplateDetailView.as_view(), name='template-detail'
    )
]



from mayan.apps.forms import form_fields

from .widgets import (
    DocumentFilePagesCarouselWidget, DocumentVersionPagesCarouselWidget,
    PageImageWidget, ThumbnailFormWidget
)


class DocumentFileField(form_fields.Field):
    widget = DocumentFilePagesCarouselWidget


class DocumentFilePageField(form_fields.Field):
    widget = PageImageWidget


class DocumentVersionField(form_fields.Field):
    widget = DocumentVersionPagesCarouselWidget


class DocumentVersionPageField(form_fields.Field):
    widget = PageImageWidget


class ThumbnailFormField(form_fields.Field):
    widget = ThumbnailFormWidget



from django.template import Library
from django.utils.safestring import mark_safe

from mayan.apps.views.http import URL

import nh3
import markdown
import requests

from ..literals import URL_FUNDRAISER_API

register = Library()


def markdown_render(source):
    md = markdown.Markdown(
        extensions=('attr_list', 'nl2br')
    )

    html = md.convert(source=source)

    html_clean = nh3.clean(
        attributes={
            'a': {'href'},
            'img': {'alt', 'class', 'src', 'style'}
        }, html=html
    )

    html_safe = mark_safe(s=html_clean)

    return html_safe


@register.simple_tag(name='fundraiser_message_fetch')
def tag_fundraiser_message_fetch(path):
    url = URL(url=URL_FUNDRAISER_API)
    url.path = path

    try:
        response = requests.get(url=url)
    except requests.exceptions.RequestException:
        return ''
    else:
        if response:
            response_json = response.json()

            body = response_json.get('body')
            title = response_json.get('title')

            body_safe = markdown_render(source=body)
            title_safe = markdown_render(source=title)

            return {
                'body': body_safe,
                'title': title_safe
            }
        else:
            return ''



from django.core.exceptions import ImproperlyConfigured
from django.views.generic.base import View

from mayan.apps.views.view_mixins import (
    MultipleObjectViewMixin, RestrictedQuerysetViewMixin, SingleObjectMixin,
    ViewPermissionCheckViewMixin
)

from .mixins import (
    ViewMixinBackendDownload, ViewMixinDownload, ViewMixinDownloadEvent
)


class ViewBaseDownload(
    ViewMixinDownload, ViewPermissionCheckViewMixin, View
):
    def get(self, request, *args, **kwargs):
        return self.render_to_response()


class ViewMultipleObjectDownload(
    RestrictedQuerysetViewMixin, MultipleObjectViewMixin, ViewBaseDownload
):
    """
    View that support receiving multiple objects via a pk_list query.
    """
    def __init__(self, *args, **kwargs):
        result = super().__init__(*args, **kwargs)

        if self.__class__.mro()[0].get_queryset != ViewMultipleObjectDownload.get_queryset:
            raise ImproperlyConfigured(
                '%(cls)s is overloading the `get_queryset` method. '
                'Subclasses should implement the `get_source_queryset` '
                'method instead. ' % {
                    'cls': self.__class__.__name__
                }
            )

        return result

    def get_queryset(self):
        try:
            return super().get_queryset()
        except ImproperlyConfigured:
            self.queryset = self.get_source_queryset()
            return super().get_queryset()


class ViewSingleObjectDownload(
    ViewMixinDownloadEvent, RestrictedQuerysetViewMixin, SingleObjectMixin,
    ViewBaseDownload
):
    def get(self, request, *args, **kwargs):
        self.object = self.get_object()
        return super().get(request=request, *args, **kwargs)

    def get_download_file_object(self):
        return self.object.open(mode='rb')

    def get_download_filename(self):
        return str(self.object)

    def get_download_event_target(self):
        return self.object


class ViewSingleObjectBackendDownload(
    ViewMixinBackendDownload, ViewSingleObjectDownload
):
    """
    Combined view to perform downloads of ORM objects, filtered by access,
    and using a download backend.
    """



import logging

from rest_framework import status

from mayan.apps.converter.api_view_mixins import APIImageViewMixin
from mayan.apps.rest_api import generics

from ..classes import DocumentVersionModification
from ..permissions import (
    permission_document_version_create, permission_document_version_delete,
    permission_document_version_edit, permission_document_version_view
)
from ..serializers.document_version_serializers import (
    DocumentVersionModificationSerializer,
    DocumentVersionModificationExecuteSerializer, DocumentVersionSerializer,
    DocumentVersionPageSerializer
)

from .api_view_mixins import (
    ParentObjectDocumentAPIViewMixin, ParentObjectDocumentVersionAPIViewMixin
)

logger = logging.getLogger(name=__name__)


class APIDocumentVersionDetailView(
    ParentObjectDocumentAPIViewMixin, generics.RetrieveUpdateDestroyAPIView
):
    """
    delete: Delete the selected document version.
    get: Returns the selected document version details.
    patch: Edit the properties of the selected document version.
    put: Edit the properties of the selected document version.
    """
    lookup_url_kwarg = 'document_version_id'
    mayan_object_permission_map = {
        'DELETE': permission_document_version_delete,
        'GET': permission_document_version_view,
        'PATCH': permission_document_version_edit,
        'PUT': permission_document_version_edit
    }
    serializer_class = DocumentVersionSerializer

    def get_instance_extra_data(self):
        return {'_event_actor': self.request.user}

    def get_source_queryset(self):
        document = self.get_document()
        return document.versions.all()


class APIDocumentVersionListView(
    ParentObjectDocumentAPIViewMixin, generics.ListCreateAPIView
):
    """
    get: Return a list of the selected document's versions.
    post: Create a new document version.
    """
    mayan_object_permission_map = {
        'GET': permission_document_version_view,
    }
    serializer_class = DocumentVersionSerializer

    def get_instance_extra_data(self):
        # This method is only called during POST, therefore filter only by
        # edit permission.
        return {
            '_event_actor': self.request.user,
            'document': self.get_document(
                permission=permission_document_version_create
            )
        }

    def get_source_queryset(self):
        document = self.get_document()
        return document.versions.all()


class APIDocumentVersionModificationView(
    ParentObjectDocumentAPIViewMixin, generics.ObjectActionAPIView
):
    """
    post: Execute a modification backend on the selected document version.
    """
    action_response_status = status.HTTP_202_ACCEPTED
    lookup_url_kwarg = 'document_version_id'
    mayan_object_permission_map = {'POST': permission_document_version_edit}
    serializer_class = DocumentVersionModificationExecuteSerializer

    def get_source_queryset(self):
        document = self.get_document()
        return document.versions.all()

    def object_action(self, obj, request, serializer):
        document_version_modification = DocumentVersionModification.get(
            name=serializer.validated_data['backend_id']
        )

        document_version_modification.execute(
            document_version=obj, user=request.user
        )


class APIDocumentVersionModificationBackendListView(generics.ListAPIView):
    """
    get: Returns a list of the available document version modification backends.
    """
    serializer_class = DocumentVersionModificationSerializer

    def get_serializer_context(self):
        return {
            'format': self.format_kwarg,
            'request': self.request,
            'view': self
        }

    def get_source_queryset(self):
        return DocumentVersionModification.get_all()


class APIDocumentVersionPageDetailView(
    ParentObjectDocumentVersionAPIViewMixin,
    generics.RetrieveUpdateDestroyAPIView
):
    """
    delete: Delete the selected document version page.
    get: Returns the selected document version page details.
    patch: Edit the properties of the selected document version page.
    put: Edit the properties of the selected document version page.
    """
    lookup_url_kwarg = 'document_version_page_id'
    mayan_object_permission_map = {
        'DELETE': permission_document_version_edit,
        'GET': permission_document_version_view,
        'PATCH': permission_document_version_edit,
        'PUT': permission_document_version_edit
    }
    serializer_class = DocumentVersionPageSerializer

    def get_instance_extra_data(self):
        return {'_event_actor': self.request.user}

    def get_source_queryset(self):
        document_version = self.get_document_version()
        return document_version.pages.all()


class APIDocumentVersionPageImageView(
    APIImageViewMixin, ParentObjectDocumentVersionAPIViewMixin,
    generics.RetrieveAPIView
):
    """
    get: Returns an image representation of the selected document version page.
    """
    lookup_url_kwarg = 'document_version_page_id'
    mayan_object_permission_map = {'GET': permission_document_version_view}

    def get_source_queryset(self):
        document_version = self.get_document_version()
        return document_version.pages.all()


class APIDocumentVersionPageListView(
    ParentObjectDocumentVersionAPIViewMixin, generics.ListCreateAPIView
):
    """
    get: Returns an list of the pages for the selected document version.
    post: Create a new document version page.
    """
    lookup_url_kwarg = 'document_version_page_id'
    serializer_class = DocumentVersionPageSerializer

    def get_instance_extra_data(self):
        # This method is only called during POST, therefore filter only by
        # edit permission.
        return {
            '_event_actor': self.request.user,
            'document_version': self.get_document_version(
                permission=permission_document_version_edit
            )
        }

    def get_source_queryset(self):
        # This method is only called during GET, therefore filter only by
        # the view permission.
        document_version = self.get_document_version(
            permission=permission_document_version_view
        )
        return document_version.pages.all()



from django.db import migrations

from mayan.apps.databases.literals import (
    DJANGO_POSITIVE_INTEGER_FIELD_MAX_VALUE
)


def code_document_file_size_update(apps, schema_editor):
    """
    Only update the document files that where set to the exact maximum field
    value in migration 0080. Set the actual size of the stored file now that
    the field allows bigger values.
    """
    DocumentFile = apps.get_model(
        app_label='documents', model_name='DocumentFile'
    )

    queryset = DocumentFile.objects.filter(
        size=DJANGO_POSITIVE_INTEGER_FIELD_MAX_VALUE
    )

    for document_file in queryset:
        name = document_file.file.name
        document_file.file.close()

        if document_file.file.storage.exists(name=name):
            document_file.size = document_file.file.storage.size(name=name)
            document_file.save(
                update_fields=('size',)
            )


def reverse_code_document_file_size_update(apps, schema_editor):
    DocumentFile = apps.get_model(
        app_label='documents', model_name='DocumentFile'
    )

    queryset = DocumentFile.objects.filter(
        size__gt=DJANGO_POSITIVE_INTEGER_FIELD_MAX_VALUE
    )

    for document_file in queryset:
        name = document_file.file.name
        document_file.file.close()

        if document_file.file.storage.exists(name=name):
            document_file.size = DJANGO_POSITIVE_INTEGER_FIELD_MAX_VALUE
            document_file.save(
                update_fields=('size',)
            )


class Migration(migrations.Migration):
    dependencies = [
        ('documents', '0081_alter_documentfile_size')
    ]

    operations = [
        migrations.RunPython(
            code=code_document_file_size_update,
            reverse_code=reverse_code_document_file_size_update
        )
    ]



import json
import logging
from pathlib import Path

import sh

from django.utils.translation import gettext_lazy as _

from mayan.apps.file_metadata.classes import FileMetadataDriver
from mayan.apps.storage.utils import TemporaryDirectory

from .literals import DEFAULT_EXIF_PATH

logger = logging.getLogger(name=__name__)


class FileMetadataDriverEXIF(FileMetadataDriver):
    argument_name_list = ('exiftool_path',)
    description = _(message='Read meta information stored in files.')
    dotted_path_previous_list = (
        'mayan.apps.file_metadata.drivers.exiftool.EXIFToolDriver',
    )
    internal_name = 'exiftool'
    label = _(message='EXIF Tool')
    mime_type_list = ('*',)

    @classmethod
    def get_argument_values_from_settings(cls):
        result = {'exiftool_path': DEFAULT_EXIF_PATH}

        setting_arguments = super().get_argument_values_from_settings()

        if setting_arguments:
            result.update(setting_arguments)

        return result

    def __init__(self, exiftool_path, **kwargs):
        super().__init__(**kwargs)

        self.command_exiftool = sh.Command(path=exiftool_path)
        self.command_exiftool = self.command_exiftool.bake('-j')

    def _process(self, document_file):
        if self.command_exiftool:
            with TemporaryDirectory() as temporary_folder:
                path_temporary_file = Path(
                    temporary_folder, Path(document_file.filename).name
                )

                with path_temporary_file.open(mode='xb') as temporary_fileobject:
                    document_file.save_to_file(
                        file_object=temporary_fileobject
                    )
                    temporary_fileobject.seek(0)
                    try:
                        output = self.command_exiftool(
                            temporary_fileobject.name
                        )
                    except sh.ErrorReturnCode_1 as exception:
                        result = json.loads(s=exception.stdout)[0]
                        if result.get('Error', '') == 'Unknown file type':
                            # Not a fatal error.
                            return result
                    else:
                        return json.loads(s=output)[0]
        else:
            logger.error(
                'EXIFTool binary not found, not processing document '
                'file: %s', document_file
            )



from django.template import Library
from django.utils.safestring import mark_safe

from mayan.apps.common.serialization import yaml_dump

register = Library()


@register.filter(name='yaml_dump')
def tag_yaml_dump(value):
    """
    Converts the given value into a YAML-formatted string.
    """

    # Convert value from SafeString to normal string.
    data = value.strip()

    result = yaml_dump(data=data).strip()

    result_safe = mark_safe(s=result)

    return result_safe



from django.db import models
from django.utils.translation import gettext_lazy as _

from mayan.apps.databases.model_mixins import (
    ExtraDataModelMixin, ModelMixinConditionField
)
from mayan.apps.events.decorators import method_event
from mayan.apps.events.event_managers import (
    EventManagerMethodAfter, EventManagerSave
)

from ..events import event_workflow_template_edited

from .workflow_models import Workflow
from .workflow_state_models import WorkflowState
from .workflow_transition_model_mixins import (
    WorkflowTransitionBusinessLogicMixin
)

__all__ = ('WorkflowTransition',)


class WorkflowTransition(
    ExtraDataModelMixin, ModelMixinConditionField,
    WorkflowTransitionBusinessLogicMixin, models.Model
):
    _condition_help_text = _(
        message='The condition that will determine if this transition '
        'is enabled or not. The condition is evaluated against the '
        'workflow instance. Conditions that do not return any value, '
        'that return the Python logical None, or an empty string (\'\') '
        'are considered to be logical false, any other value is '
        'considered to be the logical true.'
    )
    _ordering_fields = ('label',)

    workflow = models.ForeignKey(
        on_delete=models.CASCADE, related_name='transitions', to=Workflow,
        verbose_name=_(message='Workflow')
    )
    label = models.CharField(
        help_text=_(message='A short text to describe the transition.'),
        max_length=255, verbose_name=_(message='Label')
    )
    origin_state = models.ForeignKey(
        on_delete=models.CASCADE, related_name='origin_transitions',
        to=WorkflowState, verbose_name=_(message='Origin state')
    )
    destination_state = models.ForeignKey(
        on_delete=models.CASCADE, related_name='destination_transitions',
        to=WorkflowState, verbose_name=_(message='Destination state')
    )

    class Meta:
        ordering = ('label',)
        unique_together = (
            'workflow', 'label', 'origin_state', 'destination_state'
        )
        verbose_name = _(message='Workflow transition')
        verbose_name_plural = _(message='Workflow transitions')

    def __str__(self):
        return self.label

    @method_event(
        action_object='self',
        event_manager_class=EventManagerMethodAfter,
        event=event_workflow_template_edited,
        target='workflow'
    )
    def delete(self, *args, **kwargs):
        return super().delete(*args, **kwargs)

    @method_event(
        event_manager_class=EventManagerSave,
        created={
            'action_object': 'self',
            'event': event_workflow_template_edited,
            'target': 'workflow'
        },
        edited={
            'action_object': 'self',
            'event': event_workflow_template_edited,
            'target': 'workflow'
        }
    )
    def save(self, *args, **kwargs):
        return super().save(*args, **kwargs)



from mayan.apps.icons.icons import Icon

icon_message_create = icon_tag_create = Icon(
    driver_name='fontawesome-dual', primary_symbol='comment-alt',
    secondary_symbol='plus'
)
icon_message_delete = Icon(driver_name='fontawesome', symbol='times')
icon_message_detail = Icon(driver_name='fontawesome', symbol='comment-alt')
icon_message_list = Icon(driver_name='fontawesome', symbol='comment-alt')
icon_message_mark_read = Icon(driver_name='fontawesome', symbol='eye')
icon_message_mark_read_all = Icon(driver_name='fontawesome', symbol='eye')
icon_message_mark_unread = Icon(
    driver_name='fontawesome', symbol='eye-slash'
)



from mayan.apps.icons.icons import Icon

icon_color_random = Icon(driver_name='fontawesome', symbol='random')

icon_dropzone_fail = Icon(driver_name='fontawesome', symbol='times-circle')
icon_dropzone_file = Icon(
    driver_name='fontawesome-css', css_classes='far fa-file fa-10x'
)
icon_dropzone_ok = Icon(driver_name='fontawesome', symbol='check-circle')
icon_dropzone_upload = Icon(
    driver_name='fontawesome', symbol='cloud-upload-alt'
)

icon_fail = Icon(driver_name='fontawesome', symbol='times')
icon_ok = Icon(driver_name='fontawesome', symbol='check')



from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.urls import reverse
from django.utils.translation import gettext_lazy as _

from mayan.apps.databases.model_mixins import ExtraDataModelMixin
from mayan.apps.events.decorators import method_event
from mayan.apps.events.event_managers import (
    EventManagerMethodAfter, EventManagerSave
)

from ..events import (
    event_document_version_page_created, event_document_version_page_deleted,
    event_document_version_page_edited
)
from ..managers import ValidDocumentVersionPageManager

from .document_version_models import DocumentVersion
from .document_version_page_model_mixins import (
    DocumentVersionPageBusinessLogicMixin
)
from .model_mixins import PagedModelMixin

__all__ = ('DocumentVersionPage', 'DocumentVersionPageSearchResult')


class DocumentVersionPage(
    DocumentVersionPageBusinessLogicMixin, ExtraDataModelMixin,
    PagedModelMixin, models.Model
):
    _paged_model_parent_field = 'document_version'

    document_version = models.ForeignKey(
        on_delete=models.CASCADE, related_name='version_pages',
        to=DocumentVersion, verbose_name=_(message='Document version')
    )
    page_number = models.PositiveIntegerField(
        db_index=True, default=1, help_text=_(
            message='Unique integer number for the page. Pages are ordered '
            'by this number.'
        ), verbose_name=_(message='Page number')
    )
    content_type = models.ForeignKey(
        help_text=_(
            message='Content type for the source object of the page.'
        ), on_delete=models.CASCADE, to=ContentType
    )
    object_id = models.PositiveIntegerField(
        help_text=_(message='ID for the source object of the page.')
    )
    content_object = GenericForeignKey(
        ct_field='content_type', fk_field='object_id'
    )

    class Meta:
        ordering = ('page_number',)
        unique_together = ('document_version', 'page_number')
        verbose_name = _(message='Document version page')
        verbose_name_plural = _(message='Document version pages')

    objects = models.Manager()
    valid = ValidDocumentVersionPageManager()

    def __str__(self):
        return self.get_label()

    @method_event(
        event_manager_class=EventManagerMethodAfter,
        event=event_document_version_page_deleted,
        target='document_version'
    )
    def delete(self, *args, **kwargs):
        self.cache_partition.delete()
        super().delete(*args, **kwargs)

    def get_absolute_url(self):
        return reverse(
            kwargs={'document_version_page_id': self.pk},
            viewname='documents:document_version_page_view'
        )

    @method_event(
        event_manager_class=EventManagerSave,
        created={
            'action_object': 'document_version',
            'event': event_document_version_page_created,
            'target': 'self'
        },
        edited={
            'action_object': 'document_version',
            'event': event_document_version_page_edited,
            'target': 'self'
        }
    )
    def save(self, *args, **kwargs):
        return super().save(*args, **kwargs)


class DocumentVersionPageSearchResult(DocumentVersionPage):
    class Meta:
        proxy = True
        verbose_name = _(message='Document version page')
        verbose_name_plural = _(message='Document version pages')



from rest_framework import status
from rest_framework.response import Response

from mayan.apps.source_stored_files.source_backend_actions.arguments import (
    argument_file_cleanup, argument_file_identifier
)
from mayan.apps.source_stored_files.source_backend_actions.mixins import (
    SourceBackendActionMixinFileStoredBase
)
from mayan.apps.sources.source_backend_actions.interfaces import (
    SourceBackendActionInterface, SourceBackendActionInterfaceRequestRESTAPI,
    SourceBackendActionInterfaceRequestView, SourceBackendActionInterfaceTask
)


class SourceBackendActionMixinFileGenerated(
    SourceBackendActionMixinFileStoredBase
):
    class Interface:
        class Model(SourceBackendActionInterface):
            def process_interface_context(self):
                super().process_interface_context()

                self.action_kwargs['file_identifier'] = ''

        class RESTAPI(SourceBackendActionInterfaceRequestRESTAPI):
            def process_action_data(self):
                super().process_action_data()

                self.interface_result = Response(
                    status=status.HTTP_202_ACCEPTED
                )

            def process_interface_context(self):
                super().process_interface_context()

                self.action_kwargs['file_identifier'] = ''

        class Task(SourceBackendActionInterfaceTask):
            class Argument:
                file_cleanup = argument_file_cleanup
                file_identifier = argument_file_identifier

            def process_interface_context(self):
                super().process_interface_context()

                self.action_kwargs['file_cleanup'] = True
                self.action_kwargs['file_identifier'] = ''

        class View(SourceBackendActionInterfaceRequestView):
            def process_interface_context(self):
                super().process_interface_context()

                self.action_kwargs['file_identifier'] = ''



from pathlib import Path

from django.core.files.storage import FileSystemStorage
from django.utils._os import safe_join


class ShardedDirectoryFileSystemStorageMixin:
    DEFAULT_SHARDING_LEVELS = 2

    def __init__(self, *args, sharding_levels=None, **kwargs):
        if sharding_levels is None:
            sharding_levels = self.DEFAULT_SHARDING_LEVELS

        self.sharding_levels = sharding_levels

        super().__init__(*args, **kwargs)

    def get_sharded_name(self, name):
        path_original = Path(name)

        filename = path_original.name

        parts = [
            filename[:level] for level in range(1, self.sharding_levels + 1)
        ]

        path_new = Path(*parts, filename)

        return path_new

    def get_sharded_path(self, name):
        sharded_name = self.get_sharded_name(name=name)

        result = safe_join(self.location, sharded_name)

        return result


class ShardedDirectoryFileSystemStorage(
    ShardedDirectoryFileSystemStorageMixin, FileSystemStorage
):
    def _save(self, name, content):
        get_sharded_name = self.get_sharded_name(name=name)
        return super()._save(content=content, name=get_sharded_name)



from django.utils.translation import gettext_lazy as _

from mayan.apps.forms import forms


class CabinetListForm(forms.FilteredSelectionForm):
    class Meta:
        allow_multiple = True
        field_name = 'cabinets'
        label = _(message='Cabinets')
        required = False
        widget_attributes = {'class': 'select2'}



from django.shortcuts import get_object_or_404

from mayan.apps.documents.models.document_models import Document
from mayan.apps.documents.permissions import permission_document_view
from mayan.apps.documents.serializers.document_serializers import (
    DocumentSerializer
)
from mayan.apps.rest_api import generics
from mayan.apps.rest_api.api_view_mixins import ExternalObjectAPIViewMixin

from ..models.index_instance_models import IndexInstance
from ..permissions import permission_index_instance_view
from ..serializers.index_instance_serializers import (
    IndexInstanceNodeSerializer, IndexInstanceSerializer
)

from .index_instance_api_view_mixins import APIIndexInstanceNodeViewMixin


class APIDocumentIndexInstanceNodeListView(
    ExternalObjectAPIViewMixin, generics.ListAPIView
):
    """
    get: Returns a list of all the indexes instance nodes where this
    document is found.
    """
    external_object_pk_url_kwarg = 'document_id'
    external_object_queryset = Document.valid.all()
    mayan_external_object_permission_map = {
        'GET': permission_index_instance_view
    }
    mayan_object_permission_map = {'GET': permission_index_instance_view}
    serializer_class = IndexInstanceNodeSerializer

    def get_source_queryset(self):
        return self.get_external_object().index_instance_nodes.all()


class APIIndexInstanceDetailView(generics.RetrieveAPIView):
    """
    get: Returns the details of the selected index instance.
    """
    lookup_url_kwarg = 'index_instance_id'
    mayan_object_permission_map = {'GET': permission_index_instance_view}
    serializer_class = IndexInstanceSerializer
    source_queryset = IndexInstance.objects.all()


class APIIndexInstanceListView(generics.ListAPIView):
    """
    get: Returns a list of all the indexes instances.
    """
    mayan_object_permission_map = {'GET': permission_index_instance_view}
    serializer_class = IndexInstanceSerializer
    source_queryset = IndexInstance.objects.all()


class APIIndexInstanceNodeListView(
    APIIndexInstanceNodeViewMixin, generics.ListAPIView
):
    """
    get: Returns a list of all the nodes for the selected index instance.
    """
    def get_source_queryset(self):
        return self.get_index_instance().get_children()


class APIIndexInstanceNodeDetailView(
    APIIndexInstanceNodeViewMixin, generics.RetrieveAPIView
):
    """
    get: Returns the details of the selected index instance node.
    """
    lookup_url_kwarg = 'index_instance_node_id'

    def get_source_queryset(self):
        return self.get_index_instance().get_descendants()


class APIIndexInstanceNodeChildrenNodeListView(
    APIIndexInstanceNodeViewMixin, generics.ListAPIView
):
    """
    get: Returns list of all the children nodes for the selected index
    instance node.
    """
    lookup_url_kwarg = 'index_instance_node_id'

    def get_node(self):
        return get_object_or_404(
            klass=self.get_index_instance().get_descendants(),
            pk=self.kwargs['index_instance_node_id']
        )

    def get_source_queryset(self):
        return self.get_node().get_children()


class APIIndexInstanceNodeDocumentListView(
    APIIndexInstanceNodeViewMixin, generics.ListAPIView
):
    """
    get: Returns a list of all the documents contained by a particular
    index instance node.
    """
    mayan_object_permission_map = {'GET': permission_document_view}
    serializer_class = DocumentSerializer

    def get_node(self):
        return get_object_or_404(
            klass=self.get_index_instance().get_descendants(),
            pk=self.kwargs['index_instance_node_id']
        )

    def get_source_queryset(self):
        return Document.valid.filter(
            pk__in=self.get_node().documents.values('pk')
        )



from django.core import management
from django.utils.translation import gettext_lazy as _

from ...classes import Dependency


class Command(management.BaseCommand):
    help = 'Output the status of the defined dependencies.'

    def add_arguments(self, parser):
        parser.add_argument(
            '--csv', action='store_true', dest='csv', help=_(
                message='Outputs the dependencies as a comma delimited values list.'
            ),
        )

    def handle(self, *args, **options):
        Dependency.check_all(as_csv=options['csv'], use_color=True)



import logging

from django.core.files.base import ContentFile
from django.db import models
from django.template.defaultfilters import filesizeformat
from django.utils.translation import gettext_lazy as _, gettext

logger = logging.getLogger(name=__name__)


class ModelMixinFileFieldOpen:
    def open(self, **kwargs):
        # Some storage class do not provide a file mode attribute.
        # In that case default to read only in binary mode.
        # Python's default is read only in text format which does not work
        # for this use case.
        # https://docs.python.org/3/library/functions.html#open

        field_file = getattr(self, 'file')

        # Ensure the caller cannot specify an alternate filename.
        name = kwargs.pop('name', None)

        if name:
            logger.critical(
                'Caller specified an alternate filename "%s" for '
                'object: %s. Verify that this is not an attempt '
                'to access unauthorized files.', name, self
            )

        try:
            getattr(field_file.file, 'mode', None)
        except AttributeError:
            # Storage does not support mode. Discard caller supplied mode.
            kwargs.pop('mode', None)

        name = field_file.name

        open_kwargs = {'name': name}

        open_kwargs.update(**kwargs)

        # Close the self.file object as Django generates a new descriptor
        # when the file field is accessed.
        # From django/db/models/fields/files.py.
        """
        The descriptor for the file attribute on the model instance. Return a
        FieldFile when accessed so you can write code like::

            >>> from myapp.models import MyModel
            >>> instance = MyModel.objects.get(pk=1)
            >>> instance.file.size

        Assign a file object on assignment so you can do::

            >>> with open('/path/to/hello.world', 'r') as f:
            ...     instance.file = File(f)
        """
        field_file.close()
        field_file.file.close()

        return self._open(**open_kwargs)


class DatabaseFileModelMixin(ModelMixinFileFieldOpen, models.Model):
    filename = models.CharField(
        db_index=True, max_length=255, verbose_name=_(message='Filename')
    )
    datetime = models.DateTimeField(
        auto_now_add=True, verbose_name=_(message='Date time')
    )

    class Meta:
        abstract = True

    def _open(self, **kwargs):
        return self.file.storage.open(**kwargs)

    def delete(self, *args, **kwargs):
        name = self.file.name
        self.file.close()

        if name:
            self.file.storage.delete(name=name)
        return super().delete(*args, **kwargs)

    def save(self, *args, **kwargs):
        if not self.file:
            self.file = ContentFile(
                content=b'', name=self.filename or gettext(
                    message='Unnamed file'
                )
            )

        self.filename = self.filename or str(self.file)
        super().save(*args, **kwargs)


class DownloadFileBusinessLogicMixin:
    def get_size_display(self):
        return filesizeformat(bytes_=self.file.size)

    get_size_display.short_description = _(message='Size')

    def get_user_display(self):
        if self.user.get_full_name():
            return self.user.get_full_name()
        else:
            return self.user.username
    get_user_display.short_description = _(message='User')



import os

from django.conf import settings
from django.utils.translation import gettext_lazy as _

DEFAULT_GRAPHVIZ_DOT_PATH = '/usr/bin/dot'

DEFAULT_HTTP_ACTION_TIMEOUT = 4  # 4 seconds.
DEFAULT_HTTP_ACTION_VERIFY_CERTIFICATES = True

DEFAULT_WORKFLOWS_IMAGE_CACHE_MAXIMUM_SIZE = 50 * 2 ** 20  # 50 Megabytes.
DEFAULT_WORKFLOWS_IMAGE_CACHE_STORAGE_BACKEND = 'django.core.files.storage.FileSystemStorage'
DEFAULT_WORKFLOWS_IMAGE_CACHE_STORAGE_BACKEND_ARGUMENTS = {
    'location': os.path.join(settings.MEDIA_ROOT, 'workflows')
}

DEFAULT_WORKFLOWS_WORKFLOW_STATE_ESCALATION_CHECK_INTERVAL = 60 * 5  # 5 minutes.

ERROR_LOG_DOMAIN_NAME = 'workflows'

FIELD_TYPE_CHOICE_CHAR = 1
FIELD_TYPE_CHOICE_INTEGER = 2
FIELD_TYPE_CHOICES = (
    (FIELD_TYPE_CHOICE_CHAR, _(message='Character')),
    (FIELD_TYPE_CHOICE_INTEGER, _(message='Number (Integer)'))
)
FIELD_TYPE_MAPPING = {
    FIELD_TYPE_CHOICE_CHAR: 'django.forms.CharField',
    FIELD_TYPE_CHOICE_INTEGER: 'django.forms.IntegerField'
}

GRAPHVIZ_COLOR_STATE_FILL_FINAL = 'dimgray'
GRAPHVIZ_COLOR_STATE_FILL_INITIAL = 'gainsboro'
GRAPHVIZ_COLOR_STATE_FONT_COLOR_FINAL = 'white'
GRAPHVIZ_ID_STATE = 'state'
GRAPHVIZ_ID_STATE_ACTION = 'stateaction'
GRAPHVIZ_RANKDIR = 'LR'
GRAPHVIZ_RANKSEP = '1.1'
GRAPHVIZ_SHAPE_CIRCLE = 'circle'
GRAPHVIZ_SHAPE_DOUBLECIRCLE = 'doublecircle'
GRAPHVIZ_SPLINES = 'polyline'
GRAPHVIZ_STYLE_FILLED = 'filled'
GRAPHVIZ_SYMBOL_CONDITIONAL = '(&fnof;)'
GRAPHVIZ_SYMBOL_TIME = '(&tau;)'
GRAPHVIZ_SYMBOL_TRIGGER = '(!)'

STORAGE_NAME_WORKFLOW_CACHE = 'document_states__workflowimagecache'

WIDGET_CLASS_TEXTAREA = 1
WIDGET_CLASS_CHOICES = (
    (WIDGET_CLASS_TEXTAREA, _(message='Text area')),
)
WIDGET_CLASS_MAPPING = {
    WIDGET_CLASS_TEXTAREA: 'django.forms.widgets.Textarea'
}

WORKFLOW_ACTION_ON_ENTRY = 1
WORKFLOW_ACTION_ON_EXIT = 2
WORKFLOW_ACTION_WHEN_CHOICES = (
    (WORKFLOW_ACTION_ON_ENTRY, _(message='On entry')),
    (WORKFLOW_ACTION_ON_EXIT, _(message='On exit'))
)

WORKFLOW_ACTION_HTTP_REQUEST_DEFAULT_RESPONSE_STORE_NAME = 'last_http_request'



from mayan.apps.smart_settings.setting_namespaces import (
    SettingNamespaceMigration
)
from mayan.apps.smart_settings.utils import smart_yaml_load


class OCRSettingMigration(SettingNamespaceMigration):
    def ocr_backend_0002(self, value):
        """
        The PyOCR backed was removed in version 3.5, this migration
        switches the backend to the Tesseract one.
        """
        return 'mayan.apps.ocr.backends.tesseract.Tesseract'

    def ocr_backend_arguments_0001(self, value):
        """
        From version 0001 to 0002 backend arguments are no longer quoted
        but YAML valid too. Changed in version 3.3.
        """
        return smart_yaml_load(value=value)



from django.urls import reverse
from django.utils.translation import gettext_lazy as _

from mayan.apps.user_management.permissions import permission_user_view
from mayan.apps.user_management.querysets import get_user_queryset
from mayan.apps.user_management.views.view_mixins import (
    DynamicExternalUserViewMixin
)

from .generics import MultipleObjectDeleteView, SingleObjectListView
from .icons import (
    icon_user_confirm_property_view, icon_user_confirm_property_view_delete,
    icon_user_view_modes
)
from .models import UserConfirmView
from .view_mixins import ExternalObjectViewMixin


class UserConfirmViewDeleteView(MultipleObjectDeleteView):
    error_message = _(
        message='Error deleting confirm view "%(instance)s"; %(exception)s'
    )
    model = UserConfirmView
    pk_url_kwarg = 'user_confirmation_view_id'
    success_message_plural = _(
        message='%(count)d confirm view properties deleted successfully.'
    )
    success_message_single = _(
        message='Confirm view property "%(object)s" deleted successfully.'
    )
    success_message_singular = _(
        message='%(count)d confirm view property deleted successfully.'
    )
    title_plural = _(
        message='Delete the %(count)d selected confirm view property'
    )
    title_single = _(message='Delete confirm view property: %(object)s')
    title_singular = _(
        message='Delete the %(count)d selected confirm view property'
    )
    view_icon = icon_user_confirm_property_view_delete

    def get_post_action_redirect(self):
        return reverse(
            kwargs={'user_id': self.request.user.pk},
            viewname='views:user_views_confirm_list'
        )


class UserConfirmViewListView(
    DynamicExternalUserViewMixin, ExternalObjectViewMixin,
    SingleObjectListView
):
    external_object_permission = permission_user_view
    external_object_pk_url_kwarg = 'user_id'
    view_icon = icon_user_confirm_property_view

    def get_external_object_queryset(self):
        return get_user_queryset(user=self.request.user)

    def get_extra_context(self):
        return {
            'hide_link': True,
            'hide_object': True,
            'no_results_icon': icon_user_confirm_property_view,
            'no_results_text': _(
                message='Persistent confirmation properties modes remember '
                'the user preferences for a given confirmation view.'
            ),
            'no_results_title': _(
                message='No confirmation view properties available'
            ),
            'object': self.external_object,
            'title': _(
                message='Confirmation view properties for user: %s'
            ) % self.external_object
        }

    def get_source_queryset(self):
        return self.external_object.view_confirms.all()


class UserViewModeView(
    DynamicExternalUserViewMixin, ExternalObjectViewMixin,
    SingleObjectListView
):
    external_object_permission = permission_user_view
    external_object_pk_url_kwarg = 'user_id'
    view_icon = icon_user_view_modes

    def get_external_object_queryset(self):
        return get_user_queryset(user=self.request.user)

    def get_extra_context(self):
        return {
            'hide_link': True,
            'hide_object': True,
            'no_results_icon': icon_user_view_modes,
            'no_results_text': _(
                message='View modes control the format used to display a '
                'collection of objects.'
            ),
            'no_results_title': _(message='No view modes available'),
            'object': self.external_object,
            'title': _(
                message='Persistent view modes for user: %s'
            ) % self.external_object
        }

    def get_source_queryset(self):
        return self.external_object.view_modes.all()



from django.utils.translation import gettext_lazy as _

from mayan.apps.databases.classes import ModelField, ModelFieldRelated
from mayan.apps.documents.models.document_models import Document
from mayan.apps.forms import form_fields, forms
from mayan.apps.templating.fields import ModelTemplateField

from .models import SmartLink, SmartLinkCondition


class SmartLinkForm(forms.ModelForm):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields['dynamic_label'] = ModelTemplateField(
            initial_help_text=self.fields['dynamic_label'].help_text,
            label=self.fields['dynamic_label'].label, model=Document,
            model_variable='document', required=False
        )

    class Meta:
        fields = ('label', 'dynamic_label', 'enabled')
        model = SmartLink


class SmartLinkConditionForm(forms.ModelForm):
    fieldsets = (
        (
            _(message='General'), {
                'fields': ('enabled', 'inclusion',)
            },
        ),
        (
            _(message='Foreign document'), {
                'fields': ('foreign_document_data',)
            },
        ),
        (
            _(message='Local document'), {
                'fields': ('operator', 'expression', 'negated')
            },
        ),
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        choices = []
        choices.append(
            (
                ModelField.class_label, ModelField.get_choices_for(
                    model=Document
                )
            )
        )
        choices.append(
            (
                ModelFieldRelated.class_label,
                ModelFieldRelated.get_choices_for(
                    model=Document,
                )
            )
        )

        self.fields['foreign_document_data'] = form_fields.ChoiceField(
            choices=choices, label=_(message='Foreign document field')
        )
        self.fields['foreign_document_data'].widget.attrs = {
            'class': 'form-control select2-templating'
        }
        self.fields['expression'] = ModelTemplateField(
            initial_help_text=self.fields['expression'].help_text,
            label=self.fields['expression'].label, model=Document,
            model_variable='document', required=False
        )
        self.fields['inclusion'].widget.attrs = {
            'class': 'form-control select2'
        }
        self.fields['operator'].widget.attrs = {
            'class': 'form-control select2'
        }

    class Meta:
        exclude = ('smart_link',)
        model = SmartLinkCondition



from django.utils.text import format_lazy
from django.utils.translation import gettext_lazy as _

import mayan
from mayan.apps.forms import form_fields

from .template_backends import TemplateContextEntry
from .widgets import ModelTemplateWidget, TemplateWidget


class TemplateField(form_fields.CharField):
    widget = TemplateWidget

    def __init__(
        self, context_entry_name_list=None, initial_help_text='', *args,
        **kwargs
    ):
        context_entry_name_list = context_entry_name_list or ()
        self.initial_help_text = initial_help_text

        super().__init__(*args, **kwargs)

        format_string = '{initial_help_text} {template_help_text}'

        format_kwargs = {
            'initial_help_text': self.initial_help_text,
            'template_help_text': _(
                message='Use Django\'s default templating language. '
            ) % {
                'django_version': mayan.__django_version__
            }
        }

        context_variable_help_text = self.get_context_variable_help_text(
            context_entry_name_list=context_entry_name_list
        )

        if context_variable_help_text:
            format_string = '{initial_help_text} {template_help_text} {available_variable_help_text}'

            format_kwargs['available_variable_help_text'] = _(
                message='Available template context variables: %s'
            ) % context_variable_help_text

        self.help_text = format_lazy(
            format_string=format_string, **format_kwargs
        )

    def get_context_variable_help_text(self, context_entry_name_list):
        return TemplateContextEntry.get_as_help_text(
            entry_name_list=context_entry_name_list
        )


class ModelTemplateField(TemplateField):
    widget = ModelTemplateWidget

    def __init__(
        self, model, model_variable, model_variable_help_text=None, *args,
        **kwargs
    ):
        self.model = model
        self.model_variable = model_variable
        self.model_variable_help_text = model_variable_help_text

        super().__init__(*args, **kwargs)

        self.widget.attrs['app_label'] = self.model._meta.app_label
        self.widget.attrs['data-model-variable'] = self.model_variable
        self.widget.attrs['model_name'] = self.model._meta.model_name

    def get_context_variable_help_text(self, **kwargs):
        model_verbose_name = getattr(self.model._meta, 'verbose_name', None)

        model_variable_help_text = self.model_variable_help_text or model_verbose_name

        if model_variable_help_text:
            format_variable_help_text = '{{{{ {model_variable} }}}} - {model_variable_help_text}'
        else:
            format_variable_help_text = '{{{{ {model_variable} }}}}'

        variable_help_text = format_variable_help_text.format(
            model_variable=self.model_variable,
            model_variable_help_text=model_variable_help_text
        )

        result = '{}, {}'.format(
            super().get_context_variable_help_text(**kwargs),
            variable_help_text
        )

        return result



from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group
from django.contrib.auth.password_validation import validate_password
from django.utils.translation import gettext_lazy as _

from mayan.apps.rest_api import serializers
from mayan.apps.rest_api.relations import FilteredPrimaryKeyRelatedField

from .permissions import permission_user_edit
from .querysets import get_user_queryset


class GroupSerializer(serializers.HyperlinkedModelSerializer):
    users_url = serializers.HyperlinkedIdentityField(
        help_text=_(
            message='URL of the API endpoint showing the list users of this '
            'group.'
        ), label=_(message='Users URL'), lookup_url_kwarg='group_id',
        view_name='rest_api:group-user-list'
    )
    users_add_url = serializers.HyperlinkedIdentityField(
        label=_(message='Users add URL'), lookup_url_kwarg='group_id',
        view_name='rest_api:group-user-add'
    )
    users_remove_url = serializers.HyperlinkedIdentityField(
        label=_(message='Users remove URL'), lookup_url_kwarg='group_id',
        view_name='rest_api:group-user-remove'
    )

    class Meta:
        extra_kwargs = {
            'url': {
                'label': _(message='URL'),
                'lookup_url_kwarg': 'group_id',
                'view_name': 'rest_api:group-detail'
            }
        }
        fields = (
            'id', 'name', 'url', 'users_url', 'users_add_url',
            'users_remove_url'
        )
        model = Group


class GroupUserAddSerializer(serializers.Serializer):
    user = FilteredPrimaryKeyRelatedField(
        help_text=_(
            message='Primary key of the user to add to the group.'
        ), label=_(message='User ID'), source_permission=permission_user_edit,
        source_queryset=get_user_queryset()
    )


class GroupUserRemoveSerializer(serializers.Serializer):
    user = FilteredPrimaryKeyRelatedField(
        help_text=_(
            message='Primary key of the user to remove from the group.'
        ), label=_(message='User ID'), source_permission=permission_user_edit,
        source_queryset=get_user_queryset()
    )


class UserSerializer(serializers.HyperlinkedModelSerializer):
    groups_url = serializers.HyperlinkedIdentityField(
        help_text=_(
            message='URL of the API endpoint showing the list groups this '
            'user belongs to.'
        ), label=_(message='Groups URL'), lookup_url_kwarg='user_id',
        view_name='rest_api:user-group-list'
    )
    password = serializers.CharField(
        label=_(message='Password'), required=False, style={
            'input_type': 'password'
        }, write_only=True
    )

    class Meta:
        extra_kwargs = {
            'url': {
                'label': _(message='URL'),
                'lookup_url_kwarg': 'user_id',
                'view_name': 'rest_api:user-detail'
            }
        }
        fields = (
            'first_name', 'date_joined', 'email', 'groups_url',
            'id', 'is_active', 'last_login', 'last_name', 'password', 'url',
            'username'
        )
        model = get_user_model()
        read_only_fields = ('is_active', 'last_login', 'date_joined')
        write_only_fields = ('password',)

    def create(self, validated_data):
        password = validated_data.pop('password', None)
        instance = super().create(validated_data)

        if password:
            instance._event_ignore = True
            instance.set_password(raw_password=password)
            instance.save()

        return instance

    def update(self, instance, validated_data):
        if 'password' in validated_data:
            instance.set_password(
                raw_password=validated_data['password']
            )
            validated_data.pop('password')

        return super().update(
            instance=instance, validated_data=validated_data
        )

    def validate(self, data):
        if 'password' in data:
            validate_password(
                password=data['password'], user=self.instance
            )

        return data



from django.template import RequestContext
from django.urls import reverse_lazy
from django.utils.translation import gettext_lazy as _

from mayan.apps.databases.classes import ModelQueryFields
from mayan.apps.documents.models.document_models import Document
from mayan.apps.documents.permissions import permission_document_view
from mayan.apps.documents.views.document_views import DocumentListView
from mayan.apps.views.generics import (
    MultipleObjectDeleteView, SingleObjectCreateView, SingleObjectEditView,
    SingleObjectListView
)
from mayan.apps.views.view_mixins import ExternalObjectViewMixin

from ..forms import TagForm
from ..icons import (
    icon_menu_tags, icon_tag_create, icon_tag_delete_single,
    icon_tag_document_list, icon_tag_edit, icon_tag_list
)
from ..links import link_tag_create
from ..models import Tag
from ..permissions import (
    permission_tag_create, permission_tag_delete, permission_tag_edit,
    permission_tag_view
)


class TagCreateView(SingleObjectCreateView):
    extra_context = {
        'title': _(message='Create tag')
    }
    form_class = TagForm
    post_action_redirect = reverse_lazy(viewname='tags:tag_list')
    view_icon = icon_tag_create
    view_permission = permission_tag_create

    def get_instance_extra_data(self):
        return {'_event_actor': self.request.user}


class TagDeleteView(MultipleObjectDeleteView):
    error_message = _(
        message='Error deleting tag "%(instance)s"; %(exception)s'
    )
    model = Tag
    object_permission = permission_tag_delete
    pk_url_kwarg = 'tag_id'
    post_action_redirect = reverse_lazy(viewname='tags:tag_list')
    success_message_plural = _(message='%(count)d tags deleted successfully.')
    success_message_single = _(
        message='Tag "%(object)s" deleted successfully.'
    )
    success_message_singular = _(
        message='%(count)d tag deleted successfully.'
    )
    title_plural = _(message='Delete the %(count)d selected tags')
    title_single = _(message='Delete tag: %(object)s')
    title_singular = _(message='Delete the %(count)d selected tag')
    view_icon = icon_tag_delete_single

    def get_extra_context(self):
        context = super().get_extra_context()
        context = {
            'message': _(message='Will be removed from all documents.')
        }

        return context


class TagEditView(SingleObjectEditView):
    form_class = TagForm
    model = Tag
    object_permission = permission_tag_edit
    pk_url_kwarg = 'tag_id'
    view_icon = icon_tag_edit

    def get_extra_context(self):
        return {
            'object': self.object,
            'title': _(message='Edit tag: %s') % self.object
        }

    def get_instance_extra_data(self):
        return {'_event_actor': self.request.user}


class TagListView(SingleObjectListView):
    object_permission = permission_tag_view
    tag_model = Tag
    view_icon = icon_tag_list

    def get_extra_context(self):
        return {
            'hide_link': True,
            'hide_object': True,
            'no_results_icon': icon_menu_tags,
            'no_results_main_link': link_tag_create.resolve(
                context=RequestContext(request=self.request)
            ),
            'no_results_text': _(
                message='Tags are color coded properties that can be attached or '
                'removed from documents.'
            ),
            'no_results_title': _(message='No tags available'),
            'title': _(message='Tags')
        }

    def get_source_queryset(self):
        queryset = ModelQueryFields.get(model=self.tag_model).get_queryset()
        return queryset.filter(
            pk__in=self.get_tag_queryset()
        )

    def get_tag_queryset(self):
        return Tag.objects.all()


class TagDocumentListView(ExternalObjectViewMixin, DocumentListView):
    external_object_class = Tag
    external_object_permission = permission_tag_view
    external_object_pk_url_kwarg = 'tag_id'
    view_icon = icon_tag_document_list

    def get_document_queryset(self):
        return Document.valid.filter(
            pk__in=self.get_tag().get_documents(
                permission=permission_document_view,
                user=self.request.user
            ).values('pk')
        )

    def get_extra_context(self):
        context = super().get_extra_context()
        context.update(
            {
                'object': self.get_tag(),
                'title': _(
                    message='Documents with the tag: %s'
                ) % self.get_tag()
            }
        )
        return context

    def get_tag(self):
        return self.external_object



from mayan.apps.rest_api import generics

from .api_view_mixins import APIViewMixinNotification


class APINotificationDetailView(
    APIViewMixinNotification, generics.RetrieveUpdateDestroyAPIView
):
    """
    delete: Delete the selected notification.
    get: Return the details of the selected notification.
    patch: Edit the selected notification.
    put: Edit the selected notification.
    """
    lookup_url_kwarg = 'notification_id'


class APINotificationListView(APIViewMixinNotification, generics.ListAPIView):
    """
    get: Return a list of notifications for the current user.
    """



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('document_signatures', '0012_auto_20200917_0605')
    ]

    operations = [
        migrations.RemoveField(
            model_name='signaturebasemodel',
            name='date',
        ),
        migrations.AddField(
            model_name='signaturebasemodel',
            name='date_time',
            field=models.DateTimeField(
                blank=True, editable=False, null=True,
                verbose_name='Date and time signed'
            )
        )
    ]



from django.utils.translation import gettext_lazy as _

from mayan.apps.documents.permissions import (
    permission_document_tools, permission_document_view
)
from mayan.apps.navigation.links import Link

from .icons import (
    icon_duplicated_document_list, icon_duplicated_document_scan
)

link_document_duplicates_list = Link(
    args='resolved_object.id', icon=icon_duplicated_document_list,
    permission=permission_document_view, text=_(message='Duplicates'),
    view='duplicates:document_duplicates_list',
)
link_duplicated_document_list = Link(
    icon=icon_duplicated_document_list, text=_(message='Duplicates'),
    view='duplicates:duplicated_document_list'
)
link_duplicated_document_scan = Link(
    icon=icon_duplicated_document_scan,
    permission=permission_document_tools,
    text=_(message='Duplicated document scan'),
    view='duplicates:duplicated_document_scan'
)



from django.utils.translation import gettext_lazy as _

from mayan.apps.forms import form_fields, form_widgets, forms

from .literals import MATCH_ALL_FIELD_CHOICES, MATCH_ALL_FIELD_NAME
from .search_models import SearchModel
from .settings import setting_store_results_default_value


class SearchFormBase(forms.Form):
    _save_results = form_fields.BooleanField(
        help_text=_(
            message='Store the search results to speed up paging and for '
            'later browsing.'
        ), initial=setting_store_results_default_value.value, required=False,
        label=_(message='Save results')
    )

    def get_fieldsets(self):
        return (
            (
                _(message='Persistency'), {
                    'fields': ('_save_results',)
                },
            ),
        )


class AdvancedSearchForm(SearchFormBase):
    def __init__(self, *args, **kwargs):
        kwargs['data'] = kwargs['data'].copy()
        self.search_model = kwargs.pop('search_model')
        super().__init__(*args, **kwargs)

        fieldsets_dict = {}

        self.fields[MATCH_ALL_FIELD_NAME] = form_fields.ChoiceField(
            choices=MATCH_ALL_FIELD_CHOICES, label=_(message='Match all'),
            help_text=_(message='Return only results that match all fields.'),
            required=False, widget=form_widgets.RadioSelect
        )

        for search_field in self.search_model.search_fields_label_sorted:
            if search_field.concrete and not search_field.field_name == 'id':
                self.fields[search_field.field_name] = form_fields.CharField(
                    help_text=search_field.get_help_text(),
                    label=search_field.label, required=False,
                    widget=form_widgets.TextInput(
                        attrs={'type': 'search'}
                    )
                )

                # Build the fieldset dictionary.
                model = search_field.field_name_model_list[0]

                try:
                    search_model = SearchModel.get_for_model(instance=model)
                    model_verbose_name_plural = search_model.label
                except KeyError:
                    model_verbose_name_plural = model._meta.verbose_name_plural

                fieldsets_dict.setdefault(
                    model_verbose_name_plural, {
                        'fields': []
                    }
                )
                fieldsets_dict[model_verbose_name_plural]['fields'].append(
                    search_field.field_name
                )

        # Convert the fieldset dictionary to the standard fieldset tuple.
        fieldsets = ()

        fieldsets += (
            (
                _(message='Search logic'), {
                    'fields': (MATCH_ALL_FIELD_NAME,)
                }
            ),
        )

        keys = list(
            fieldsets_dict.keys()
        )
        keys.sort()

        for key in keys:
            fieldsets += (
                key, fieldsets_dict[key]
            ),

        self._fieldsets = fieldsets

    def get_fieldsets(self):
        fieldsets = super().get_fieldsets()

        fieldsets += self._fieldsets

        return fieldsets


class SearchForm(SearchFormBase):
    q = form_fields.CharField(
        max_length=128, label=_(message='Search terms'), required=False,
        widget=form_widgets.TextInput(
            attrs={'type': 'search'}
        )
    )

    def get_fieldsets(self):
        fieldsets_super = super().get_fieldsets()

        fieldsets = (
            (
                _(message='Criteria'), {
                    'fields': ('q',)
                },
            ),
        )

        fieldsets += fieldsets_super

        return fieldsets



from django.utils.translation import gettext_lazy as _

from .literals import (
    DEFAULT_ALLOWED_HOSTS, DEFAULT_APPEND_SLASH,
    DEFAULT_AUTH_PASSWORD_VALIDATORS, DEFAULT_AUTHENTICATION_BACKENDS,
    DEFAULT_CACHES, DEFAULT_CSRF_COOKIE_SECURE, DEFAULT_CSRF_TRUSTED_ORIGINS,
    DEFAULT_CSRF_USE_SESSIONS, DEFAULT_DATA_UPLOAD_MAX_MEMORY_SIZE,
    DEFAULT_DATABASES, DEFAULT_DEFAULT_FROM_EMAIL,
    DEFAULT_DISALLOWED_USER_AGENTS, DEFAULT_EMAIL_BACKEND, DEFAULT_EMAIL_HOST,
    DEFAULT_EMAIL_HOST_PASSWORD, DEFAULT_EMAIL_HOST_USER, DEFAULT_EMAIL_PORT,
    DEFAULT_EMAIL_TIMEOUT, DEFAULT_EMAIL_USE_SSL, DEFAULT_EMAIL_USE_TLS,
    DEFAULT_FILE_UPLOAD_MAX_MEMORY_SIZE, DEFAULT_INTERNAL_IPS,
    DEFAULT_LANGUAGE_CODE, DEFAULT_LANGUAGES, DEFAULT_LOGIN_REDIRECT_URL,
    DEFAULT_LOGIN_URL, DEFAULT_LOGOUT_REDIRECT_URL,
    DEFAULT_SECURE_PROXY_SSL_HEADER, DEFAULT_SESSION_COOKIE_NAME,
    DEFAULT_SESSION_ENGINE, DEFAULT_STATIC_URL, DEFAULT_STORAGES,
    DEFAULT_TIME_ZONE, DEFAULT_USE_X_FORWARDED_HOST,
    DEFAULT_USE_X_FORWARDED_PORT, DEFAULT_WSGI_APPLICATION
)
from .setting_clusters import SettingCluster

# Don't import anything on star imports, we just want to make it easy
# for apps.py to activate the settings in this module.
__all__ = ()
setting_cluster = SettingCluster(name='primary')
setting_namespace = setting_cluster.do_namespace_add(
    label=_(message='Django'), name='django'
)

setting_django_allowed_hosts = setting_namespace.do_setting_add(
    default=DEFAULT_ALLOWED_HOSTS, global_name='ALLOWED_HOSTS', help_text=_(
        message='A list of strings representing the host/domain names that this site '
        'can serve. This is a security measure to prevent HTTP Host header '
        'attacks, which are possible even under many seemingly-safe web '
        'server configurations. Values in this list can be '
        'fully qualified names (e.g. \'www.example.com\'), in which case '
        'they will be matched against the request\'s Host header exactly '
        '(case-insensitive, not including port). A value beginning with a '
        'period can be used as a subdomain wildcard: \'.example.com\' will '
        'match example.com, www.example.com, and any other subdomain of '
        'example.com. A value of \'*\' will match anything; in this case you '
        'are responsible to provide your own validation of the Host header '
        '(perhaps in a middleware; if so this middleware must be listed '
        'first in MIDDLEWARE).'
    )
)
setting_django_append_slash = setting_namespace.do_setting_add(
    default=DEFAULT_APPEND_SLASH, global_name='APPEND_SLASH', help_text=_(
        message='When set to True, if the request URL does not match any of the '
        'patterns in the URLconf and it doesn\'t end in a slash, an HTTP '
        'redirect is issued to the same URL with a slash appended. Note '
        'that the redirect may cause any data submitted in a POST request '
        'to be lost. The APPEND_SLASH setting is only used if '
        'CommonMiddleware is installed (see Middleware). See also '
        'PREPEND_WWW.'
    )
)
setting_django_auth_password_validators = setting_namespace.do_setting_add(
    default=DEFAULT_AUTH_PASSWORD_VALIDATORS,
    global_name='AUTH_PASSWORD_VALIDATORS', help_text=_(
        message='The list of validators that are used to check the strength of '
        'user\'s passwords.'
    )
)
setting_django_authentication_backends = setting_namespace.do_setting_add(
    default=DEFAULT_AUTHENTICATION_BACKENDS,
    global_name='AUTHENTICATION_BACKENDS', help_text=_(
        message='A list of authentication backend classes (as strings) to use when '
        'attempting to authenticate a user.'
    )
)
setting_django_caches = setting_namespace.do_setting_add(
    default=DEFAULT_CACHES,
    global_name='CACHES', help_text=_(
        message='A dictionary containing the settings for all caches to be '
        'used with Django. It is a nested dictionary whose contents maps '
        'cache aliases to a dictionary containing the options for an '
        'individual cache.'
    )
)
setting_django_csrf_cookie_secure = setting_namespace.do_setting_add(
    default=DEFAULT_CSRF_COOKIE_SECURE, global_name='CSRF_COOKIE_SECURE',
    help_text=_(
        message='Whether to use a secure cookie for the CSRF cookie. '
        'If this is set to True, the cookie will be marked as "secure", '
        'which means browsers may ensure that the cookie is only sent '
        'with an HTTPS connection. '
        'https://docs.djangoproject.com/en/4.2/ref/settings/#std-setting-CSRF_COOKIE_SECURE'
    )
)
setting_django_csrf_trusted_origins = setting_namespace.do_setting_add(
    default=DEFAULT_CSRF_TRUSTED_ORIGINS, global_name='CSRF_TRUSTED_ORIGINS',
    help_text=_(
        message='A list of trusted origins for unsafe requests (e.g. POST). '
        'https://docs.djangoproject.com/en/4.2/ref/settings/#std-setting-CSRF_TRUSTED_ORIGINS'
    )
)
setting_django_csrf_use_sessions = setting_namespace.do_setting_add(
    default=DEFAULT_CSRF_USE_SESSIONS, global_name='CSRF_USE_SESSIONS',
    help_text=_(
        message='Whether to store the CSRF token in the user\'s session '
        'instead of in a cookie. '
        'https://docs.djangoproject.com/en/4.2/ref/settings/#std-setting-CSRF_USE_SESSIONS'
    )
)
setting_django_databases = setting_namespace.do_setting_add(
    default=DEFAULT_DATABASES, global_name='DATABASES', help_text=_(
        message='A dictionary containing the settings for all databases to be used '
        'with Django. It is a nested dictionary whose contents map a '
        'database alias to a dictionary containing the options for an '
        'individual database. The DATABASES setting must configure a '
        'default database; any number of additional databases may also '
        'be specified.'
    )
)
setting_django_data_upload_max_memory_size = setting_namespace.do_setting_add(
    default=DEFAULT_DATA_UPLOAD_MAX_MEMORY_SIZE,
    global_name='DATA_UPLOAD_MAX_MEMORY_SIZE', help_text=_(
        message='Default: 2621440 (i.e. 2.5 MB). The maximum size in bytes that a '
        'request body may be before a SuspiciousOperation '
        '(RequestDataTooBig) is raised. The check is done when accessing '
        'request.body or request.POST and is calculated against the total '
        'request size excluding any file upload data. You can set this to '
        'None to disable the check. Applications that are expected to '
        'receive unusually large form posts should tune this setting. The '
        'amount of request data is correlated to the amount of memory '
        'needed to process the request and populate the GET and POST '
        'dictionaries. Large requests could be used as a '
        'denial-of-service attack vector if left unchecked. Since web '
        'servers don\'t typically perform deep request inspection, it\'s '
        'not possible to perform a similar check at that level. See also '
        'FILE_UPLOAD_MAX_MEMORY_SIZE.'
    )
)
setting_django_default_from_email = setting_namespace.do_setting_add(
    default=DEFAULT_DEFAULT_FROM_EMAIL, global_name='DEFAULT_FROM_EMAIL',
    help_text=_(
        message='Default: \'webmaster@localhost\' '
        'Default email address to use for various automated correspondence '
        'from the site manager(s). This doesn\'t include error messages sent '
        'to ADMINS and MANAGERS; for that, see SERVER_EMAIL.'
    )
)
setting_django_disallowed_user_agents = setting_namespace.do_setting_add(
    default=DEFAULT_DISALLOWED_USER_AGENTS,
    global_name='DISALLOWED_USER_AGENTS', help_text=_(
        message='Default: [] (Empty list). List of compiled regular expression '
        'objects representing User-Agent strings that are not allowed to '
        'visit any page, systemwide. Use this for bad robots/crawlers. '
        'This is only used if CommonMiddleware is installed '
        '(see Middleware).'
    )
)
setting_django_email_backend = setting_namespace.do_setting_add(
    default=DEFAULT_EMAIL_BACKEND, global_name='EMAIL_BACKEND', help_text=_(
        message='Default: \'django.core.mail.backends.smtp.EmailBackend\'. The '
        'backend to use for sending emails.'
    )
)
setting_django_email_host = setting_namespace.do_setting_add(
    default=DEFAULT_EMAIL_HOST, global_name='EMAIL_HOST', help_text=_(
        message='Default: \'localhost\'. The host to use for sending email.'
    )
)
setting_django_email_host_password = setting_namespace.do_setting_add(
    default=DEFAULT_EMAIL_HOST_PASSWORD, global_name='EMAIL_HOST_PASSWORD',
    help_text=_(
        message='Default: \'\' (Empty string). Password to use for the SMTP '
        'server defined in EMAIL_HOST. This setting is used in '
        'conjunction with EMAIL_HOST_USER when authenticating to the '
        'SMTP server. If either of these settings is empty, '
        'Django won\'t attempt authentication.'
    )
)
setting_django_email_host_user = setting_namespace.do_setting_add(
    default=DEFAULT_EMAIL_HOST_USER, global_name='EMAIL_HOST_USER',
    help_text=_(
        message='Default: \'\' (Empty string). Username to use for the SMTP '
        'server defined in EMAIL_HOST. If empty, Django won\'t attempt '
        'authentication.'
    )
)
setting_django_email_port = setting_namespace.do_setting_add(
    default=DEFAULT_EMAIL_PORT, global_name='EMAIL_PORT', help_text=_(
        message='Default: 25. Port to use for the SMTP server defined in EMAIL_HOST.'
    )
)
setting_django_email_timeout = setting_namespace.do_setting_add(
    default=DEFAULT_EMAIL_TIMEOUT, global_name='EMAIL_TIMEOUT', help_text=_(
        message='Default: None. Specifies a timeout in seconds for blocking '
        'operations like the connection attempt.'
    )
)
setting_django_email_user_ssl = setting_namespace.do_setting_add(
    default=DEFAULT_EMAIL_USE_SSL, global_name='EMAIL_USE_SSL', help_text=_(
        message='Default: False. Whether to use an implicit TLS (secure) connection '
        'when talking to the SMTP server. In most email documentation this '
        'type of TLS connection is referred to as SSL. It is generally used '
        'on port 465. If you are experiencing problems, see the explicit '
        'TLS setting EMAIL_USE_TLS. Note that EMAIL_USE_TLS/EMAIL_USE_SSL '
        'are mutually exclusive, so only set one of those settings to True.'
    )
)
setting_django_email_user_tls = setting_namespace.do_setting_add(
    default=DEFAULT_EMAIL_USE_TLS, global_name='EMAIL_USE_TLS', help_text=_(
        message='Default: False. Whether to use a TLS (secure) connection when '
        'talking to the SMTP server. This is used for explicit TLS '
        'connections, generally on port 587. If you are experiencing '
        'hanging connections, see the implicit TLS setting EMAIL_USE_SSL.'
    )
)
setting_django_file_upload_max_memory_size = setting_namespace.do_setting_add(
    default=DEFAULT_FILE_UPLOAD_MAX_MEMORY_SIZE,
    global_name='FILE_UPLOAD_MAX_MEMORY_SIZE', help_text=_(
        message='Default: 2621440 (i.e. 2.5 MB). The maximum size (in bytes) '
        'that an upload will be before it gets streamed to the file '
        'system. See Managing files for details. See also '
        'DATA_UPLOAD_MAX_MEMORY_SIZE.'
    )
)
setting_django_login_url = setting_namespace.do_setting_add(
    default=DEFAULT_LOGIN_URL, global_name='LOGIN_URL', help_text=_(
        message='Default: \'/accounts/login/\' The URL where requests are '
        'redirected for login, especially when using the login_required() '
        'decorator. This setting also accepts named URL patterns which '
        'can be used to reduce configuration duplication since you '
        'don\'t have to define the URL in two places (settings '
        'and URLconf).'
    )
)
setting_django_login_redirect_url = setting_namespace.do_setting_add(
    default=DEFAULT_LOGIN_REDIRECT_URL, global_name='LOGIN_REDIRECT_URL',
    help_text=_(
        message='Default: \'/accounts/profile/\' The URL where requests are '
        'redirected after login when the contrib.auth.login view gets no '
        'next parameter. This is used by the login_required() decorator, '
        'for example. This setting also accepts named URL patterns which '
        'can be used to reduce configuration duplication since you don\'t '
        'have to define the URL in two places (settings and URLconf).'
    )
)
setting_django_logout_redirect_url = setting_namespace.do_setting_add(
    default=DEFAULT_LOGOUT_REDIRECT_URL, global_name='LOGOUT_REDIRECT_URL',
    help_text=_(
        message='Default: None. The URL where requests are redirected after a user '
        'logs out using LogoutView (if the view doesn\'t get a next_page '
        'argument). If None, no redirect will be performed and the logout '
        'view will be rendered. This setting also accepts named URL '
        'patterns which can be used to reduce configuration duplication '
        'since you don\'t have to define the URL in two places (settings '
        'and URLconf).'
    )
)
setting_django_internal_ips = setting_namespace.do_setting_add(
    default=DEFAULT_INTERNAL_IPS, global_name='INTERNAL_IPS', help_text=_(
        message='A list of IP addresses, as strings, that: Allow the debug() '
        'context processor to add some variables to the template context. '
        'Can use the admindocs bookmarklets even if not logged in as a '
        'staff user. Are marked as "internal" (as opposed to "EXTERNAL") '
        'in AdminEmailHandler emails.'
    )
)
setting_django_languages = setting_namespace.do_setting_add(
    default=DEFAULT_LANGUAGES, global_name='LANGUAGES', help_text=_(
        message='A list of all available languages. The list is a list of '
        'two-tuples in the format (language code, language name) '
        'for example, (\'ja\', \'Japanese\'). This specifies which '
        'languages are available for language selection. '
        'Generally, the default value should suffice. Only set this '
        'setting if you want to restrict language selection to a '
        'subset of the Django-provided languages. '
    )
)
setting_django_language_code = setting_namespace.do_setting_add(
    default=DEFAULT_LANGUAGE_CODE, global_name='LANGUAGE_CODE', help_text=_(
        message='A string representing the language code for this installation. '
        'This should be in standard language ID format. For example, U.S. '
        'English is "en-us". It serves two purposes: If the locale '
        'middleware isn\'t in use, it decides which translation is served '
        'to all users. If the locale middleware is active, it provides a '
        'fallback language in case the user\'s preferred language can\'t '
        'be determined or is not supported by the website. It also provides '
        'the fallback translation when a translation for a given literal '
        'doesn\'t exist for the user\'s preferred language.'
    )
)
setting_django_cookie_name = setting_namespace.do_setting_add(
    default=DEFAULT_SESSION_COOKIE_NAME, global_name='SESSION_COOKIE_NAME',
    help_text=_(
        message='Default: \'sessionid\'. The name of the cookie to use for sessions.'
        'This can be whatever you want (as long as it\'s different from the '
        'other cookie names in your application).'
    )
)
setting_django_session_engine = setting_namespace.do_setting_add(
    default=DEFAULT_SESSION_ENGINE, global_name='SESSION_ENGINE',
    help_text=_(
        message='Default: \'django.contrib.sessions.backends.db\'. Controls where '
        'Django stores session data.'
    )
)
setting_django_secure_proxy_ssl_header = setting_namespace.do_setting_add(
    default=DEFAULT_SECURE_PROXY_SSL_HEADER,
    global_name='SECURE_PROXY_SSL_HEADER', help_text=_(
        message='A tuple representing a HTTP header/value combination that '
        'signifies a request is secure. This controls the behavior of the '
        'request object’s is_secure() method. Warning: Modifying this '
        'setting can compromise your site’s security. Ensure you fully '
        'understand your setup before changing it.'
    )
)
setting_django_static_url = setting_namespace.do_setting_add(
    default=DEFAULT_STATIC_URL, global_name='STATIC_URL', help_text=_(
        message='URL to use when referring to static files located in STATIC_ROOT. '
        'Example: "/static/" or "http://static.example.com/" '
        'If not None, this will be used as the base path for asset '
        'definitions (the Media class) and the staticfiles app. '
        'It must end in a slash if set to a non-empty value.'
    )
)
setting_django_storages = setting_namespace.do_setting_add(
    default=DEFAULT_STORAGES, global_name='STORAGES',
    help_text=_(
        message='A dictionary containing the settings for all storages to be used '
        'with Django. It is a nested dictionary whose contents map a '
        'storage alias to a dictionary containing the options for an '
        'individual storage.'
    )
)
setting_django_time_zone = setting_namespace.do_setting_add(
    default=DEFAULT_TIME_ZONE, global_name='TIME_ZONE', help_text=_(
        message='A string representing the time zone for this installation. '
        'Note that this isn\'t necessarily the time zone of the server. '
        'For example, one server may serve multiple Django-powered sites, '
        'each with a separate time zone setting.'
    )
)
setting_django_wsgi_application = setting_namespace.do_setting_add(
    default=DEFAULT_USE_X_FORWARDED_HOST, global_name='USE_X_FORWARDED_HOST',
    help_text=_(
        message='A boolean that specifies whether to use the X-Forwarded-Host '
        'header in preference to the Host header. This should only be '
        'enabled if a proxy which sets this header is in use.'
    )
)
setting_django_wsgi_application = setting_namespace.do_setting_add(
    default=DEFAULT_USE_X_FORWARDED_PORT, global_name='USE_X_FORWARDED_PORT',
    help_text=_(
        message='A boolean that specifies whether to use the X-Forwarded-Port '
        'header in preference to the SERVER_PORT META variable. This '
        'should only be enabled if a proxy which sets this header is in '
        'use. USE_X_FORWARDED_HOST takes priority over this setting.'
    )
)
setting_django_wsgi_application = setting_namespace.do_setting_add(
    default=DEFAULT_WSGI_APPLICATION, global_name='WSGI_APPLICATION',
    help_text=_(
        message='The full Python path of the WSGI application object that Django\'s '
        'built-in servers (e.g. runserver) will use. The django-admin '
        'startproject management command will create a simple wsgi.py '
        'file with an application callable in it, and point this setting '
        'to that application.'
    )
)



from django.urls import re_path

from .views import (
    QueueTaskTypeListView, TaskTypeListView, WorkerListView,
    WorkerQueueListView
)

urlpatterns_queues = [
    re_path(
        route=r'^queues/(?P<queue_name>\w+)/task_types/$',
        view=QueueTaskTypeListView.as_view(), name='queue_task_type_list'
    )
]

urlpatterns_tasks = [
    re_path(
        route=r'^task_types/$', view=TaskTypeListView.as_view(),
        name='task_type_list'
    )
]

urlpatterns_workers = [
    re_path(
        route=r'^workers/$', view=WorkerListView.as_view(),
        name='worker_list'
    ),
    re_path(
        route=r'^workers/(?P<worker_name>\w+)/queues/$',
        view=WorkerQueueListView.as_view(), name='worker_queue_list'
    )
]


urlpatterns = []
urlpatterns.extend(urlpatterns_queues)
urlpatterns.extend(urlpatterns_tasks)
urlpatterns.extend(urlpatterns_workers)



from django.core.cache import caches

from rest_framework import throttling


class MayanAnonRateThrottle(throttling.AnonRateThrottle):
    cache = caches['rest_api_throttling']


class MayanUserRateThrottle(throttling.UserRateThrottle):
    cache = caches['rest_api_throttling']



from django.utils.translation import gettext_lazy as _

from mayan.apps.navigation.menus import Menu

menu_sources = Menu(
    cache_class_associations=False, label=_(message='Sources'), name='sources'
)



from django.utils.translation import gettext_lazy as _

from mayan.apps.documents.search import (
    search_model_document, search_model_document_file
)

# Document

search_model_document.add_model_field(
    field='files__source_metadata__key', label=_(
        message='Source metadata key'
    )
)
search_model_document.add_model_field(
    field='files__source_metadata__value', label=_(
        message='Source metadata value'
    )
)

# Document file

search_model_document_file.add_model_field(
    field='source_metadata__key', label=_(message='Source metadata key')
)
search_model_document_file.add_model_field(
    field='source_metadata__value', label=_(message='Source metadata value')
)



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('document_states', '0031_alter_workflowtransitionfield_required')
    ]

    operations = [
        migrations.AddField(
            model_name='workflowtransitionfield', name='lookup',
            field=models.TextField(
                blank=True, help_text='Enter a template to render. Must '
                'result in a comma delimited string.', null=True,
                verbose_name='Lookup'
            )
        )
    ]



from mayan.apps.forms import form_widgets, forms

from .models import UserLocaleProfile


class LocaleProfileForm(forms.ModelForm):
    class Meta:
        fields = ('language', 'timezone')
        model = UserLocaleProfile
        widgets = {
            'language': form_widgets.Select(
                attrs={'class': 'select2'}
            ),
            'timezone': form_widgets.Select(
                attrs={'class': 'select2'}
            )
        }


class LocaleProfileForm_view(forms.DetailForm):
    class Meta:
        fields = ('language', 'timezone')
        model = UserLocaleProfile



from django.utils.translation import gettext_lazy as _

from mayan.apps.rest_api import serializers


class AJAXTemplateSerializer(serializers.Serializer):
    hex_hash = serializers.CharField(
        label=_(message='Hex hash'), read_only=True
    )
    name = serializers.CharField(
        label=_(message='Name'), read_only=True
    )
    html = serializers.CharField(
        label=_(message='HTML'), read_only=True
    )
    url = serializers.HyperlinkedIdentityField(
        label=_(message='URL'), lookup_field='name', lookup_url_kwarg='name',
        view_name='rest_api:template-detail'
    )


class ObjectTemplateSandboxSerializer(serializers.Serializer):
    result = serializers.CharField(
        label=_(message='Result'), read_only=True
    )
    template = serializers.CharField(
        label=_(message='Template'), write_only=True
    )



from django.utils.translation import gettext_lazy as _

from mayan.apps.navigation.links import Link

from .icons import (
    icon_saved_resultset_delete_single, icon_saved_resultset_list,
    icon_saved_resultset_result_list, icon_search, icon_search_advanced,
    icon_search_again, icon_search_backend_reindex, icon_search_model_detail,
    icon_search_model_list
)
from .permissions import permission_search_tools
from .search_backends import SearchBackend


def condition_search_backend_supports_reindexing(context, resolved_object):
    search_backend_class = SearchBackend.get_class()
    return search_backend_class.feature_reindex


link_saved_resultset_list = Link(
    icon=icon_saved_resultset_list, text=_(message='Saved resultsets'),
    view='search:saved_resultset_list'
)
link_saved_resultset_delete_single = Link(
    args=('resolved_object.pk',), icon=icon_saved_resultset_delete_single,
    tags='dangerous', text=_(message='Delete'),
    view='search:saved_resultset_delete_single'
)
link_saved_resultset_result_list = Link(
    args=('resolved_object.pk',), icon=icon_saved_resultset_result_list,
    text=_(message='Results'), view='search:saved_resultset_result_list'
)

link_search = Link(
    args='search_model.full_name', icon=icon_search,
    text=_(message='Basic search'), view='search:search_simple'
)
link_search_advanced = Link(
    args='search_model.full_name', icon=icon_search_advanced,
    text=_(message='Advanced search'), view='search:search_advanced'
)
link_search_again = Link(
    args='search_model.full_name', icon=icon_search_again,
    keep_query=True, text=_(message='Search again'), view='search:search_again'
)
link_search_backend_reindex = Link(
    condition=condition_search_backend_supports_reindexing,
    icon=icon_search_backend_reindex, permission=permission_search_tools,
    text=_(message='Reindex search backend'), view='search:search_backend_reindex'
)

# Search model

link_search_model_detail = Link(
    icon=icon_search_model_detail,
    kwargs={'search_model_name': 'resolved_object.full_name'},
    text=_(message='Fields'), view='search:search_model_detail'
)
link_search_model_list = Link(
    icon=icon_search_model_list, text=_(message='Search models'),
    view='search:search_model_list'
)



from django.template import Library

from ..models import AutoAdminSingleton

register = Library()


@register.inclusion_tag(
    filename='autoadmin/credentials.html', name='autoadmin_partial'
)
def tag_autoadmin_partial():
    try:
        return {'autoadmin_properties': AutoAdminSingleton.objects.get()}
    except AutoAdminSingleton.DoesNotExist:
        return {'autoadmin_properties': None}


@register.simple_tag(name='autoadmin_properties', takes_context=True)
def tag_autoadmin_properties(context):
    try:
        context['autoadmin_properties'] = AutoAdminSingleton.objects.get()
    except AutoAdminSingleton.DoesNotExist:
        context['autoadmin_properties'] = None

    return ''



from django.core.exceptions import ValidationError
from django.db import models
from django.urls import reverse
from django.utils.module_loading import import_string
from django.utils.translation import gettext_lazy as _

from mayan.apps.common.serialization import yaml_load
from mayan.apps.common.validators import YAMLValidator
from mayan.apps.databases.model_mixins import ExtraDataModelMixin
from mayan.apps.events.decorators import method_event
from mayan.apps.events.event_managers import EventManagerSave

from ..events import event_metadata_type_created, event_metadata_type_edited
from ..managers import MetadataTypeManager

from .metadata_type_model_mixins import MetadataTypeBusinessLogicMixin


class MetadataType(
    ExtraDataModelMixin, MetadataTypeBusinessLogicMixin, models.Model
):
    """
    Model to store a type of metadata. Metadata are user defined properties
    that can be assigned a value for each document. Metadata types need to be
    assigned to a document type before they can be used.
    """
    _ordering_fields = ('label', 'name')

    name = models.CharField(
        max_length=48,
        help_text=_(
            message='Name used by other apps to reference this metadata '
            'type. Do not use python reserved words, or spaces.'
        ),
        unique=True, verbose_name=_(message='Name')
    )
    label = models.CharField(
        help_text=_(message='Short description of this metadata type.'),
        max_length=48, verbose_name=_(message='Label')
    )
    default = models.CharField(
        blank=True, max_length=128, null=True, help_text=_(
            message='Enter a template to render.'
        ), verbose_name=_(message='Default')
    )
    lookup = models.TextField(
        blank=True, null=True, help_text=_(
            message='Enter a template to render. Must result in a comma '
            'delimited string.'
        ), verbose_name=_(message='Lookup')
    )
    validation = models.CharField(
        blank=True, help_text=_(
            message='The validator will reject data entry if the value '
            'entered does not conform to the expected format.'
        ), max_length=224, verbose_name=_(message='Validator')
    )
    validation_arguments = models.TextField(
        blank=True, help_text=_(
            message='Enter the arguments for the validator in YAML format.'
        ), validators=[YAMLValidator()], verbose_name=_(
            message='Validator arguments'
        )
    )
    parser = models.CharField(
        blank=True, help_text=_(
            message='The parser will reformat the value entered to conform '
            'to the expected format.'
        ), max_length=224, verbose_name=_(message='Parser')
    )
    parser_arguments = models.TextField(
        blank=True, help_text=_(
            message='Enter the arguments for the parser in YAML format.'
        ), validators=[
            YAMLValidator()
        ], verbose_name=_(message='Parser arguments')
    )

    objects = MetadataTypeManager()

    class Meta:
        ordering = ('label',)
        verbose_name = _(message='Metadata type')
        verbose_name_plural = _(message='Metadata types')

    def __str__(self):
        return self.label

    def get_absolute_url(self):
        return reverse(
            kwargs={'metadata_type_id': self.pk},
            viewname='metadata:metadata_type_edit'
        )

    def natural_key(self):
        return (self.name,)

    @method_event(
        created={
            'event': event_metadata_type_created,
            'target': 'self',
        },
        edited={
            'event': event_metadata_type_edited,
            'target': 'self',
        },
        event_manager_class=EventManagerSave
    )
    def save(self, *args, **kwargs):
        return super().save(*args, **kwargs)

    def validate_value(self, document_type, value):
        # Check default
        if not value and self.default:
            value = self.get_default_value()

        if not value and self.get_required_for(document_type=document_type):
            raise ValidationError(
                message=_(
                    message='"%s" is required for this document type.'
                ) % self.label
            )

        if self.lookup:
            lookup_options = self.get_lookup_values()

            if value and value not in lookup_options:
                raise ValidationError(
                    message=_(
                        message='Value is not one of the provided options.'
                    )
                )

        if self.validation:
            validator_class = import_string(dotted_path=self.validation)
            validator_arguments = yaml_load(
                stream=self.validation_arguments or '{}'
            )
            validator = validator_class(**validator_arguments)
            try:
                validator.validate(value)
            except ValidationError as exception:
                raise ValidationError(
                    message=_(
                        message='Metadata type validation error; '
                        '%(exception)s'
                    ) % {
                        'exception': ','.join(exception)
                    }
                ) from exception

        if self.parser:
            parser_class = import_string(dotted_path=self.parser)
            parser_arguments = yaml_load(
                stream=self.parser_arguments or '{}'
            )
            parser = parser_class(**parser_arguments)
            value = parser.parse(value)

        return value



from mayan.apps.forms import forms
from mayan.apps.templating.fields import ModelTemplateField

from ..models.workflow_instance_models import WorkflowInstance
from ..models.workflow_transition_models import WorkflowTransition


class WorkflowTransitionForm(forms.ModelForm):
    def __init__(self, workflow, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.fields[
            'origin_state'
        ].queryset = self.fields[
            'origin_state'
        ].queryset.filter(workflow=workflow)

        self.fields[
            'destination_state'
        ].queryset = self.fields[
            'destination_state'
        ].queryset.filter(workflow=workflow)

        self.fields['condition'] = ModelTemplateField(
            initial_help_text=self.fields['condition'].help_text,
            label=self.fields['condition'].label, model=WorkflowInstance,
            model_variable='workflow_instance', required=False
        )

    class Meta:
        fields = ('label', 'origin_state', 'destination_state', 'condition')
        model = WorkflowTransition
